require ("src/cards_3d_config")
---@class CHandCard3D @单张手牌（不包括自己的手牌），可以设牌墙（竖立背面），可以设摊开
CHandCard3D=class("CHandCard3D", function() return ImageView:create() end)
function CHandCard3D:ctor(...)
    self._size={}
    self._scale=1
    self._card_value=InvalidValue
    self._visual_value=InvalidValue
    self._card_img=nil
    self._ground_path = nil
	self:init(...)
end

-- 要可以被多次初始化
---@param card_info table
---@param base_x int
---@param base_y int
---@param base_zorder int
function CHandCard3D:init(card_info,base_x,base_y,base_zorder)
    if not self._card_img then
        self._card_img=control_tools.newImg({})
        self:addChild(self._card_img)
    end

    if not self._corner_img then
        self._corner_img=control_tools.newImg({})
        self:addChild(self._corner_img)
    end

    if not self._card_mask then
        self._card_mask=ImageView:create()
        self:addChild(self._card_mask)
    end

    if card_info ~= nil then
        -- 底
        self._ground_path = card_info.ground.file
        self:loadTexture(self._ground_path,1)
        self:setPosition(ccp(base_x+card_info.ground.x, base_y+card_info.ground.y))
        self:setScale(card_info.ground.scale)
        self:setZOrder(base_zorder+card_info.zorder)

        -- 需要考虑没有牌面的情况
        if card_info.card ~= nil then
            -- 牌面
            local card_x = self:getSize().width / 2 + card_info.card.x
            local card_y = self:getSize().height / 2 + card_info.card.y
            self._card_img:setPosition(ccp(card_x, card_y))
            self._card_img:setScale(card_info.card.scale)
            self._card_path=card_info.card.file
        end

        -- 角标
        if card_info.corner then
            self.corner_x = self:getSize().width / 2 + card_info.corner.x
            self.corner_y = self:getSize().height / 2 + card_info.corner.y
            self.corner_scale = card_info.corner.scale
            self._corner_img:setPosition(ccp(self.corner_x, self.corner_y))
            self._corner_img:setScale(card_info.corner.scale)
            self._corner_path=card_info.corner.file
        end

        -- mask
        self._card_mask:loadTexture(self._ground_path,1)
        self._card_mask:setPosition(ccp(0, 0))
        self._card_mask:setScale(1)
        self._card_mask:setOpacity(100)
        self._card_mask:setZOrder(base_zorder+card_info.zorder)
        self._card_mask:setVisible(false)
        self._card_mask:runAction(CCTintTo:create(0,20,20,20))
    end

    self:setVisible(false)
end

---@param param_value byte @牌值
---@param param_corner_id int @角标，没有为0或nil，财神为1，鸡为2，后续待添加
---@param param_show_mask bool @是否显示mask
function CHandCard3D:setCardValue(param_value,param_corner_id,param_show_mask)
    self._card_value=param_value
    if param_value == InvalidValue then
        self._visual_value = InvalidValue
        self._card_img:setVisible(false)
        self:hideCorner()
        self:hideMask()
    else
        if CardExchangeTable[param_value] ~= nil then
            param_value = CardExchangeTable[param_value]
        end
        self._visual_value = param_value
        if self._card_path then
            if (self._visual_value >= 10 and self._visual_value <= 47) or (self._visual_value >= 50 and self._visual_value <= 58) then
                local img_path = string.gsub(self._card_path, "{NUM}", self._visual_value)
                self._card_img:loadTexture(img_path,1)
                self._card_img:setVisible(true)
            end
        end

        if param_corner_id and param_corner_id ~= 0 then
            self:showCorner(param_corner_id)
        else
            self:hideCorner()
        end

        if param_show_mask and param_show_mask==true then
            self:showMask()
        else
            self:hideMask()
        end
    end
end

function CHandCard3D:showMask()
    self._card_mask:setVisible(true)
end

function CHandCard3D:hideMask()
    self._card_mask:setVisible(false)
end

function CHandCard3D:showCorner(param_corner_id)
    if param_corner_id >= 1 and param_corner_id <= 7 then
        local img_path = string.gsub(self._corner_path, "{NUM}", param_corner_id)
        self._corner_img:loadTexture(img_path,1)

        self._corner_img:setVisible(true)
        self._corner_img:setPosition(ccp(self.corner_x, self.corner_y))
        self._corner_img:setScale(self.corner_scale)
    end
end

function CHandCard3D:hideCorner()
    self._corner_img:setVisible(false)
end

function CHandCard3D:getPosX()
    return self._posx
end

function CHandCard3D:getPosY()
    return self._posy
end

---@param param_value byte @牌值
---@param param_corner_id int @角标，没有为0或nil，财神为1，后续待添加
---@param param_show_mask bool @是否显示mask
function CHandCard3D:showCard(param_value,param_corner_id,param_show_mask)
    if param_value == nil then
        param_value = self:getCardValue()
    end
    self:setCardValue(param_value,param_corner_id,param_show_mask)
    self:setVisible(true)
end

function CHandCard3D:hideCard()
    self._card_value=InvalidValue
    self._visual_value=InvalidValue
    self:setVisible(false)
end

function CHandCard3D:destroy()

end

return CHandCard3D
