require ("src/cards_3d_config")
local CHandleHint3D=require("src/Controls/3D/CHandleHint3D")
---@class CHandleSelect3D @吃碰杠牌的选择条，包含1-5个按钮（吃，碰，杠，胡，听，过）
CHandleSelect3D=class("CHandleSelect3D", function() return ImageView:create() end)
local UI_SCALE = 0.8
function CHandleSelect3D:ctor(...)
    self._btn_handle_bg = nil
    self._gang_handle_hint = nil
    self._chi_handle_hint = nil
    self._on_result_func = nil
    self._view_size = nil
    self._btn_handle = {}
    self:init(...)
end

-- 初始化handle选择面板
function CHandleSelect3D:init(view_size)
    self._view_size = view_size

    if not self._btn_handle_bg then
        self._btn_handle_bg = control_tools.newImg({path="btn_handle_handle_bg.png", ctype=1})
        self:addChild(self._btn_handle_bg)
        --就乱缩
        self._btn_handle_bg:setScale(UI_SCALE)
    end
    for i=1,6 do
        self._btn_handle[i]=control_tools.newBtn({small=true,normal="btn_handle_btn_"..i..".png", ctype=1})
        self._btn_handle[i]._index=i
        self._btn_handle[i]._type=HandleType.null
        self._btn_handle[i]._value=InvalidValue
        self._btn_handle[i]:addTouchEventListener(function(param_sender,param_touchType) self:handleBtnEvent(param_sender,param_touchType) end)
        self._btn_handle[i]:setVisible(false)
        self._btn_handle[i]:setTouchEnabled(false)
        --self._btn_handle[i]:setZOrder(CardZOrder.HandleBtn)
        self._btn_handle_bg:addChild(self._btn_handle[i])
    end
    if not self._gang_handle_hint then
        self._gang_handle_hint=CHandleHint3D.new(view_size)
        self._gang_handle_hint:setController(self)
        self._gang_handle_hint:setType("gang")
        self._btn_handle_bg:addChild(self._gang_handle_hint)
    end
    if not self._chi_handle_hint then
        self._chi_handle_hint=CHandleHint3D.new(view_size)
        self._chi_handle_hint:setController(self)
        self._chi_handle_hint:setType("chi")
        self._btn_handle_bg:addChild(self._chi_handle_hint)
    end
    self:setVisible(false)
end

function CHandleSelect3D:setOnResultFunc(func_onResult)
    if func_onResult and type(func_onResult) == "function" then
        self._on_result_func = func_onResult
    end
end

-- 设置handle数据，顺序按照传入参数的顺序，如果需要“过”显示在最左边，则第一个参数就传HandleType.pass
---@param list_handle table @{{type=HandleType.b_chi_left, value=11},{type=HandleType.b_gang, value=35}...}
function CHandleSelect3D:setHandleData(list_handle)
    local gang_data={}
    local chi_data={}
    local handle_order={}
    self._gang_handle_hint:hideHint(true)
    self._chi_handle_hint:hideHint(true)
    for i=1,#self._btn_handle do
        self._btn_handle[i]:setVisible(false)
        self._btn_handle[i]:setTouchEnabled(false)
    end

    local function make_handle_order(handle_type)
        for i=1,#handle_order do
            if handle_order[i] == handle_type then
                return
            end
        end
        table.insert(handle_order, handle_type)
    end
    for i=1,#list_handle do
        local handle_type=list_handle[i].type
        local card_value=list_handle[i].value
        if handle_type==HandleType.b_peng then
            self._btn_handle[HandleButtonIndex.Peng]._type=handle_type
            self._btn_handle[HandleButtonIndex.Peng]._value=card_value
            make_handle_order(HandleButtonIndex.Peng)
        elseif handle_type==HandleType.b_chi_left or handle_type==HandleType.b_chi_mid or handle_type==HandleType.b_chi_right then
            table.insert(chi_data,{type=handle_type,value=card_value})
            make_handle_order(HandleButtonIndex.Chi)
        elseif handle_type==HandleType.b_gang or handle_type==HandleType.d_gang or handle_type==HandleType.a_gang then
            table.insert(gang_data,{type=handle_type,value=card_value})
            make_handle_order(HandleButtonIndex.Gang)
        elseif handle_type==HandleType.hu then
            self._btn_handle[HandleButtonIndex.Hu]._type=handle_type
            self._btn_handle[HandleButtonIndex.Hu]._value=InvalidValue
            make_handle_order(HandleButtonIndex.Hu)
        elseif handle_type==HandleType.pass then
            self._btn_handle[HandleButtonIndex.Pass]._type=handle_type
            self._btn_handle[HandleButtonIndex.Pass]._value=InvalidValue
            make_handle_order(HandleButtonIndex.Pass)
        elseif handle_type == HandleType.ting then
            self._btn_handle[HandleButtonIndex.Ting]._type=handle_type
            self._btn_handle[HandleButtonIndex.Ting]._value=InvalidValue
            make_handle_order(HandleButtonIndex.Ting)
        end
    end

    local btn_count=#handle_order
    if btn_count > 0 then
        local btn_width = self._btn_handle[1]:getSize().width
        local btn_height = self._btn_handle[1]:getSize().height
        local btn_start_x = - (btn_count - 1) * btn_width * 0.5
        for i=1,btn_count do
            local handle_type = handle_order[i]
            self._btn_handle[handle_type]:setPosition(ccp(btn_start_x+(i-1)*btn_width, -7))
            self._btn_handle[handle_type]:setVisible(true)
            self._btn_handle[handle_type]:setTouchEnabled(true)
        end

        self._btn_handle_bg:setVisible(true)
        self._btn_handle_bg:setScale9Enabled(true)
        if btn_count == 1 then
            self._btn_handle_bg:setCapInsets(CCRect(30,0,480,138))
        else
            self._btn_handle_bg:setCapInsets(CCRect(135,0,270,138))
        end
        self._btn_handle_bg:setSize(CCSize(144*(btn_count+1)-70,138))

        local gap_x = 0;
        if self._view_size.width/self._view_size.height > 16/9 then
            gap_x = 70;
        end

        self._btn_handle_bg:setPosition(ccp(
        self._view_size.width-self._btn_handle_bg:getSize().width*0.5*UI_SCALE - 30 - gap_x, -- 不能紧贴右边，留30像素空余
        self._btn_handle_bg:getSize().height*(1.5+0.5)*UI_SCALE  -- 0.5本身位移，1.5是高于自己本身高度1.5倍
        ))

        if #gang_data > 0 then
            if #gang_data>1 then
                local function sort_func(param_v1,param_v2)
                    return param_v1.value>param_v2.value
                end
                table.sort(gang_data,sort_func)
                for k,v in pairs(gang_data) do
                    self._gang_handle_hint:addHintData(v.type,v.value)
                end
                local x = self._btn_handle[HandleButtonIndex.Gang]:getPositionX()
                local y = self._btn_handle[HandleButtonIndex.Gang]:getPositionY()
                y = y + btn_height * 0.5
                self._gang_handle_hint:setBasePosition(ccp(x,y))
                x = self._btn_handle_bg:convertToWorldSpace(ccp(x,y)).x
                if x > self._view_size.width * 0.5 then
                    self._gang_handle_hint:setDirRight()
                else
                    self._gang_handle_hint:setDirLeft()
                end
            else
                self._btn_handle[HandleButtonIndex.Gang]._type=gang_data[1].type
                self._btn_handle[HandleButtonIndex.Gang]._value=gang_data[1].value
            end
        end

        if #chi_data > 0 then
            if #chi_data>1 then
                for k,v in pairs(chi_data) do
                    self._chi_handle_hint:addHintData(v.type,v.value)
                end
                local x = self._btn_handle[HandleButtonIndex.Chi]:getPositionX()
                local y = self._btn_handle[HandleButtonIndex.Chi]:getPositionY()
                y = y + btn_height * 0.5
                self._chi_handle_hint:setBasePosition(ccp(x,y))
                x = self._btn_handle_bg:convertToWorldSpace(ccp(x,y)).x
                if x > self._view_size.width * 0.5 then
                    self._chi_handle_hint:setDirRight()
                else
                    self._chi_handle_hint:setDirLeft()
                end
            else
                self._btn_handle[HandleButtonIndex.Chi]._type=chi_data[1].type
                self._btn_handle[HandleButtonIndex.Chi]._value=chi_data[1].value
            end
        end
    else
        self:setVisible(false)
    end
end

function CHandleSelect3D:handleBtnEvent(param_sender,param_touchType)
    if param_touchType~=_G.TOUCH_EVENT_ENDED then
        return
    end
    local send_message=true
    if param_sender._index==HandleButtonIndex.Gang then
        if self._gang_handle_hint:getHintCount()>0 then
            if self._gang_handle_hint:isVisible()==false then
                self._gang_handle_hint:showHint()
                self._chi_handle_hint:hideHint()
            else
                self._gang_handle_hint:hideHint()
            end
            send_message=false
        end
    end
    if param_sender._index==HandleButtonIndex.Chi then
        if self._chi_handle_hint:getHintCount()>0 then
            if self._chi_handle_hint:isVisible()==false then
                self._chi_handle_hint:showHint()
                self._gang_handle_hint:hideHint()
            else
                self._chi_handle_hint:hideHint()
            end
            send_message=false
        end
    end
    if send_message==true then
        if self._on_result_func then
            self._on_result_func({handle_type=param_sender._type,card_value=param_sender._value})
        end
        if LayoutConfig.auto_hide_handle_select == true then
            self:hide()
        end
    end
end

function CHandleSelect3D:touchEvent(sender, touchType, card)
    if touchType~=_G.TOUCH_EVENT_ENDED then
        return
    end
    if self._on_result_func then
        self._on_result_func({handle_type=card._type,card_value=card._message_value})
    end
    if LayoutConfig.auto_hide_handle_select == true then
        self:hide()
    end
end

function CHandleSelect3D:show()
    self:setVisible(true)
end

function CHandleSelect3D:hide()
    self._chi_handle_hint:hideHint(true)
    self._gang_handle_hint:hideHint(true)
    self:setVisible(false)
end

function CHandleSelect3D:destroy()

end

return CHandleSelect3D
