require ("src/cards_3d_config")
local CTiledCard3D=require("src/Controls/3D/CTiledCard3D")
---@class CHandledCards3D @吃碰杠牌堆（逻辑控制），具体显示交由CTiledCard3D
CHandledCards3D=class("CHandledCards3D",function() return ImageView:create() end)
function CHandledCards3D:ctor(...)
    self._handled_card={}
    self._scale=1
    self._type=HandleType.null
    self._value=InvalidValue
    self._arrow=nil
    self._arrow_path=""
    self:init(...)
end
function CHandledCards3D:init(cards_info,base_x,base_y,base_zorder)
    if not cards_info then return end
    for i=1,4,1 do
        if not self._handled_card[i] then
            self._handled_card[i] = CTiledCard3D.new(cards_info.items[i],base_x,base_y,base_zorder)
        end
    end
    if not self._arrow then
        self._arrow_path = cards_info.direct.file
        self._arrow = control_tools.newImg({})
        self._arrow:setPosition(ccp(base_x+cards_info.direct.x, base_y+cards_info.direct.y))
        self._arrow:setScale(cards_info.direct.scale)
        self._arrow:setZOrder(base_zorder+cards_info.direct.zorder)
    end
end

function CHandledCards3D:setHandledData(e_fromIndex, param_type, param_value)
    self._type=param_type
    self._value=param_value
    self._from=e_fromIndex

    self._handled_card[4]:setCardValue(InvalidValue)
    if param_type==HandleType.b_peng then
        for i=1,3,1 do
            self._handled_card[i]:setCardValue(param_value)
        end
    elseif param_type==HandleType.b_chi_left then
        self._handled_card[1]:setCardValue(param_value)
        self._handled_card[2]:setCardValue(param_value+1)
        self._handled_card[3]:setCardValue(param_value+2)
    elseif param_type==HandleType.b_chi_mid then
        self._handled_card[1]:setCardValue(param_value-1)
        self._handled_card[2]:setCardValue(param_value)
        self._handled_card[3]:setCardValue(param_value+1)
    elseif param_type==HandleType.b_chi_right then
        self._handled_card[1]:setCardValue(param_value-2)
        self._handled_card[2]:setCardValue(param_value-1)
        self._handled_card[3]:setCardValue(param_value)
    elseif param_type==HandleType.d_gang then
        local config = {}
        if cards_3d_layout._on_config_dark_gang_func then
            config = cards_3d_layout._on_config_dark_gang_func(param_value)
        end
        if config and #config == 4 then
            -- 有单独配置
            for i=1,4 do
                if config[i] == DarkGangConfig.Value then
                    self._handled_card[i]:setCardValue(param_value)
                else
                    self._handled_card[i]:showBack()
                end
            end
        else
            -- 无单独配置，使用全局默认配置
            for i=1,3,1 do
                self._handled_card[i]:showBack()
            end
            if LayoutConfig.show_card_value_on_dark_gang then
                self._handled_card[4]:setCardValue(param_value)
            else
                self._handled_card[4]:showBack()
            end
        end
    elseif param_type==HandleType.a_gang or param_type==HandleType.b_gang then
        for i=1,4,1 do
            self._handled_card[i]:setCardValue(param_value)
        end
    end

    if LayoutConfig.show_heap_arrow then
        if self._from >=1 and self._from <= 4 then
            local img_path = string.gsub(self._arrow_path, "{NUM}", self._from)
            self._arrow:loadTexture(img_path, 1)
        end
    end
end

function CHandledCards3D:showCards(corner_config, dark_card_value)
    for i=1,3,1 do
        self._handled_card[i]:showCard(nil, corner_config[self._handled_card[i]:getCardValue()], self._handled_card[i]:getCardValue() == dark_card_value)
    end
    if self._type==HandleType.b_gang or self._type==HandleType.a_gang or self._type==HandleType.d_gang then
        self._handled_card[4]:showCard(nil, corner_config[self._handled_card[4]:getCardValue()], self._handled_card[4]:getCardValue() == dark_card_value)
    else
        self._handled_card[4]:hideCard()
    end
    if LayoutConfig.show_heap_arrow and self._arrow then
        if self._from and self._from ~= -1 then
            self._arrow:setVisible(true)
        else
            self._arrow:setVisible(false)
        end
    end
end

function CHandledCards3D:hideCards()
    for i = 1, 4 do
        self._handled_card[i]:hideCard()
    end
    if self._arrow then
        self._arrow:setVisible(false)
    end
end

function CHandledCards3D:showMask(param_value)
    --for i=1,4,1 do
    --    local card_value=self._handled_card[i]:getCardValue()
    --    if card_value==param_value and card_value~=InvalidValue then
    --        self._handled_card[i]:showMask()
    --    else
    --        self._handled_card[i]:hideMask()
    --    end
    --end
end

function CHandledCards3D:hideMask()
    --for i=1,4,1 do
    --    self._handled_card[i]:hideMask()
    --end
end

function CHandledCards3D:destroy()
    for i = 1, 4 do
        self._handled_card[i]:destroy()
    end
end

return CHandledCards3D