require ("src/cards_3d_config")
local CSelfHandCard3D = require("src/Controls/3D/CSelfHandCard3D")
local CHandledCards3D = require("src/Controls/3D/CHandledCards3D")
local cards_3d_logic=require("src/cards_3d_logic")
---@class CSelfCardsController @自己手牌控制器，负责显示自己手牌和吃碰杠牌堆，包括动画，响应输入，提供callback
CSelfCardsController=class("CSelfCardsController")
function CSelfCardsController:ctor(...)
    --self._view_size = nil
    --self._view_root = nil
    --self._dark_card_value = InvalidValue
    --self._corner_config = {}
    --self._lay = false
    --self._hand_cards = {}
    --self._heap_cards = {}
    --self:init(...)
end
---@param view_root table @scene node节点
---@param view_size table @{width=1334,height=750}
---@param int_max_handcard_count int @最大手牌数，14 or 17
---@param hand_cards table @外部创建好的table，用来储存CSelfHandCard3D手牌资源数组
---@param heap_cards table @外部创建好的table，用来储存CHandledCards3D吃碰杠牌数组
function CSelfCardsController:init(view_root, view_size, int_max_handcard_count, hand_cards, heap_cards)
    self:reset()

    self._view_root = view_root
    self._view_size = view_size
    self._hand_cards = hand_cards
    self._heap_cards = heap_cards

    -- 牌资源宽度
    self._CARD_RESOURCE_WIDTH = 72
    -- 牌资源高度
    self._CARD_RESOURCE_HEIGHT = 108
    -- 牌的背景图中上下左右有一定空白，要计算这些空白
    self._CARD_RESOURCE_EDGE = 8

    -- 手牌相对于屏幕尺寸的缩放
    local view_width = view_size.height*(1334/750);
    if view_size.width/view_size.height > 18.5/9 then
        view_width = view_size.height*(18/9);
    end
    if view_width > view_size.width then view_width = view_size.width;end
    if int_max_handcard_count == 14 then
        self._HAND_CARD_SCALE = view_width / 1334.0 * 1.28
    elseif int_max_handcard_count == 17 then
        self._HAND_CARD_SCALE = view_width / 1334.0 * 1.05
    else
        self._HAND_CARD_SCALE = view_width / 1334.0 * 1.15
    end
    -- 单款手牌额外缩放，默认是1
    self._HAND_CARD_SCALE = self._HAND_CARD_SCALE * LayoutConfig.custom_self_handcard_scale
    -- 手牌宽度
    self._HAND_CARD_WIDTH = self._CARD_RESOURCE_WIDTH * self._HAND_CARD_SCALE
    -- 手牌高度
    self._HAND_CARD_HEIGHT = self._CARD_RESOURCE_HEIGHT * self._HAND_CARD_SCALE
    -- 手牌基础偏移X，手牌位置由吃碰杠牌堆决定，如果没有吃碰杠，则是这个位置
    self._HAND_CARD_BASE_X = 90 * self._HAND_CARD_SCALE + self._HAND_CARD_WIDTH * 0.5 + (view_size.width - view_width)/2
    -- 手牌基础偏移Y
    self._HAND_CARD_BASE_Y = 46 * self._HAND_CARD_SCALE
    -- 手牌空白
    self._HAND_CARD_EDGE = self._CARD_RESOURCE_EDGE * self._HAND_CARD_SCALE
    -- 抓上的牌距离手牌的空余
    self._CATCH_HAND_CARD_SPACE = 20 * self._HAND_CARD_SCALE
    -- 被点击选中的牌的缩放
    self._SELECTED_HAND_CARD_SCALE = 1.2 * self._HAND_CARD_SCALE

    -- 吃碰杠牌和手牌的间隔
    self._HEAP_HAND_SPACE = 10 * self._HAND_CARD_SCALE

    -- 吃碰杠牌相对于手牌的缩放
    self._HEAP_CARD_SCALE = 0.8 * self._HAND_CARD_SCALE
    -- 吃碰杠牌的宽度
    self._HEAP_CARD_WIDTH = self._CARD_RESOURCE_WIDTH * self._HEAP_CARD_SCALE
    -- 吃碰杠牌的高度
    self._HEAP_CARD_HEIGHT = self._CARD_RESOURCE_HEIGHT * self._HEAP_CARD_SCALE
    -- 吃碰杠牌组间的空余
    self._HEAP_SPACE = 10 * self._HAND_CARD_SCALE
    -- 吃碰杠牌组的整体偏移
    self._HEAP_BASE_X = 100 * self._HEAP_CARD_SCALE + self._HEAP_CARD_WIDTH * 0.5 + (view_size.width - view_width)/2
    self._HEAP_BASE_Y = 0 + self._HEAP_CARD_HEIGHT * 0.5
    -- 吃碰杠的箭头Y偏移
    self._HEAP_ARROW_Y = 60 * self._HEAP_CARD_SCALE
    -- 吃碰杠的箭头缩放
    self._HEAP_ARROW_SCALE = 0.7 * self._HEAP_CARD_SCALE

    -- 一个heap里每张牌的位移
    -- 吃碰杠牌空白
    self._HEAP_CARD_EDGE = self._CARD_RESOURCE_EDGE * self._HEAP_CARD_SCALE
    self._heap_card_pos = {
        ccp(0,0),
        ccp(self._HEAP_CARD_WIDTH-self._HEAP_CARD_EDGE,0),
        ccp((self._HEAP_CARD_WIDTH-self._HEAP_CARD_EDGE)*2,0),
        ccp(self._HEAP_CARD_WIDTH-self._HEAP_CARD_EDGE,22*self._HAND_CARD_SCALE),
    }
    -- 每个heap的位移
    self._heap_pos = {}
    for i=1,5 do
        self._heap_pos[i] = ccp(self._HEAP_BASE_X + (i-1)*((self._HEAP_CARD_WIDTH-self._HEAP_CARD_EDGE)*3 + self._HEAP_SPACE), self._HEAP_BASE_Y)
    end
    -- 手牌默认的起始位置，初始化时设成和吃碰杠牌堆一致，这个值会在游戏过程中被改变
    self._hand_cards_start_x = self._HAND_CARD_BASE_X
end

function CSelfCardsController:reset()
    -- 拖动手牌的起始位置
    self._touch_start_pos = nil
    -- dark列表
    self._dark_card_value = InvalidValue
    -- 角标列表
    self._corner_config = {}
    -- 是否摊开
    self._lay = false
    -- 单独摊开的手牌
    self._lay_index = 0;
    -- 点击类型(1:多选，0:正常出牌)
    self._select_type = 0;
    -- 当前点击选中的，会被放大的手牌，被拖动的手牌不算作selected
    self._selected_index = 0
    -- 多选
    self._selected_table = {};
    -- 在插牌动画过程中，手牌被锁定
    self._lock_hand_cards = false
    -- 是否开始拖动手牌
    self._drag_began = false
    self._ting_list = {}
    self._ting_da_list = {}
    self._ting_duo_list = {}
    self:setHandCardsCanTouch(false)
    self:setHandCardsCanOut(false)
    self:setHandCardsDark(false)
    self:setHandCardsTing(nil)
end

function CSelfCardsController:setSelectedIndex(int_index)
    if self._onSelectedIndexFunc then
        self._onSelectedIndexFunc(int_index)
    end
    self._selected_index = int_index
end

function CSelfCardsController:setSelectedTable(selectTable)
    table.sort(selectTable)
    self._selected_table = selectTable;
end    

function CSelfCardsController:setSelectedValue(selectValue)
    table.sort(selectValue)
    self._selected_table = self:getHandCardsIndexByValue(selectValue);
end

function CSelfCardsController:getSelectedTable()
    local tableCard = {};
    for k,v in pairs(self._selected_table) do
        local card = self._hand_cards[v]:getCardValue();
        table.insert(tableCard,card);
    end
    return tableCard;
end

function CSelfCardsController:getSelectIndexTable()
    local tableIndex = {} ;
    for k,v in pairs(self._selected_table) do
        local card = v;
        table.insert(tableIndex, card);
    end
    return tableIndex;
end

function CSelfCardsController:setSelectType(type)
    self._select_type = type;
end

function CSelfCardsController:setSelfHandJump(index,interval,high,times,lock,unlock)
    if lock == true then self:lockHandCards(); end
    local card = nil;
    if index == -1 then
        for k,v in pairs(self._hand_cards) do
            v:stopAllActions();
            v:runAction(CCJumpBy:create(interval,ccp(0,0),high,times));
            card = v;
        end
    elseif self._hand_cards[index] then 
        self._hand_cards[index]:stopAllActions();
        self._hand_cards[index]:runAction(CCJumpBy:create(interval,ccp(0,0),high,times));
        card = self._hand_cards[index];
    end 
    
    if unlock == true then
        local function jumpEnd()
            self:unlockHandCards();
            self:refreshHandCards();
        end
        card:runAction(CCSequence:createWithTwoActions(CCDelayTime:create(interval + 0.1),CCCallFunc:create(jumpEnd)));
    end
end

function CSelfCardsController:setLayIndex(index)
    self._lay_index = index;
    self:refreshHandCards();
end

-- 生成一个手牌，参数全部为默认，不根据view_size缩放
function CSelfCardsController:newHandCard()
    return CSelfHandCard3D.new()
end

-- 生成一个heap牌，参数全部为默认，不根据view_size缩放
function CSelfCardsController:newHeapCard()

end

---@param bool_lay bool @设置手牌是否摊开
function CSelfCardsController:layHandCards(bool_lay)
    self._lay = bool_lay
    self:refreshHandCards()
end

-- 设置桌上使用灰色遮罩的牌值，如果不需要，传0或nil
---@param byte_card int
function CSelfCardsController:setDarkCardValue(byte_card)
    self._dark_card_value = byte_card
end

-- 设置有角标的牌值
---@param config table @config[47]=1;config[41]=2;...
function CSelfCardsController:setCornerCardValue(config)
    self._corner_config = config
end

---@param cards table @{11,12,13,14,...}
---@param int_show_count int @显示的牌数量
function CSelfCardsController:setHandCards(cards, int_show_count)
    -- 手牌处于锁定状态，一般是由于正在播放插牌动画
    if self._lock_hand_cards then return end

    -- 首先根据吃碰杠牌堆计算手牌起始位置
    self:refreshHandCardStartPosition()

    if not int_show_count then int_show_count = #cards end
    for i = 1, int_show_count do
        self:addHandCard(cards[i], i)
    end
    for i = #self._hand_cards,int_show_count+1,-1 do
        self:removeHandCard(i)
    end
end

---@param heaps table @{{type=HandleType.b_peng, value=21, from=UserIndex.up},{type=HandleType.b_gang, value=24, from=UserIndex.left}}
function CSelfCardsController:setHeapCards(heaps)
    -- 手牌处于锁定状态，一般是由于正在播放插牌动画
    if self._lock_hand_cards then return end

    for i = 1, #heaps do
        self:addHeapCard(heaps[i].from, heaps[i].type, heaps[i].value, i)
    end
    for i = #self._heap_cards, #heaps+1, -1 do
        self:removeHeapCard(i)
    end

    -- 设置完吃碰杠堆后可能会影响手牌位置，这里要重新计算、显示
    self:refreshHandCardStartPosition()
    self:refreshHandCards()
end

function CSelfCardsController:refreshHandCardStartPosition()
    local heaps = cards_3d_logic:getCards(UserIndex.down, LayoutCardsType.heap)
    if heaps and #heaps > 0 then
        local heap = self:getHeapCard(#heaps)
        self._hand_cards_start_x = heap._handled_card[3]:getPositionX() +          -- 最后一个吃碰杠牌组的第三张
        self._HEAP_CARD_WIDTH * 0.5 +                   -- 这张牌的半宽
        self._HEAP_HAND_SPACE +                         -- 吃碰杠和手牌的基础距离
        self._HAND_CARD_WIDTH * 0.5                     -- 第一张手牌的半宽
    else
        self._hand_cards_start_x = self._HAND_CARD_BASE_X
    end
end

function CSelfCardsController:refreshHandCards()
    -- 手牌处于锁定状态，一般是由于正在播放插牌动画
    if self._lock_hand_cards then return end

    -- 取到逻辑上的所有手牌
    local cards = cards_3d_logic:getCards(UserIndex.down, LayoutCardsType.hand)

    if #cards > 0 then
        -- 把逻辑上有的牌全都显示
        for i = 1, #cards do
            self:addHandCard(cards[i], i)
        end
    end

    -- 有多余的已显示的手牌，需要隐藏
    for i = #self._hand_cards,#cards+1,-1 do
        self:removeHandCard(i)
    end

    -- 刷新听牌标记
    -- 先隐藏所有
    for _,v in pairs(self._hand_cards) do
        if v then
            v:hideTing()
        end
    end
    if self._ting_list then
        for i=1,#self._ting_list do
            if self._hand_cards[self._ting_list[i]] then
                self._hand_cards[self._ting_list[i]]:showTing(1)
            end
        end
    end
    if self._ting_da_list then
        for i=1,#self._ting_da_list do
            if self._hand_cards[self._ting_da_list[i]] then
                self._hand_cards[self._ting_da_list[i]]:showTing(2)
            end
        end
    end
    if self._ting_duo_list then
        for i=1,#self._ting_duo_list do
            if self._hand_cards[self._ting_duo_list[i]] then
                self._hand_cards[self._ting_duo_list[i]]:showTing(3)
            end
        end
    end
end

function CSelfCardsController:unlockHandCards()
    self._lock_hand_cards = false
    for i=1,#self._hand_cards do
        if self._hand_cards[i] then
            self._hand_cards[i]._touch_paused = false
        end
    end
end

function CSelfCardsController:lockHandCards()
    self._lock_hand_cards = true
    for i=1,#self._hand_cards do
        if self._hand_cards[i] then
            self._hand_cards[i]._touch_paused = true
        end
    end
end

---@param int_pos int @出掉的牌原本所在的index，会空出来
function CSelfCardsController:insertLastCardInHandCards(int_pos)
    local cards = cards_3d_logic:getCards(UserIndex.down, LayoutCardsType.hand)
    if not cards or #cards == 0 then return end

    -- 最后一张牌
    local last_index = #cards
    if int_pos == last_index then
        -- 出掉的牌是最后一张牌，其他牌都不动
        return
    end

    self:lockHandCards()

    local card = self._hand_cards[#cards]
    local card_value = cards[#cards]
    local other_cards = clone(cards)
    table.remove(other_cards, int_pos)  -- 去掉要出掉的牌
    table.remove(other_cards)   -- 去掉最后一个

    local index = #other_cards+1 -- 默认移动到最后一张
    if self._findCardInHandPosition then
        index = self._findCardInHandPosition(other_cards, card_value)
    else
        for i,v in ipairs(other_cards) do
            if v > card_value then
                index = i
                break
            end
        end
    end

    local function move_end_func()
        -- 移完以后要重新设logic里的位置
        --cards_3d_logic:moveCard(UserIndex.down, LayoutCardsType.hand, #cards, index)
        cards_3d_logic:sortHandCards(UserIndex.down)
        self:unlockHandCards()
        self:refreshHandCards()
        if self._onInsertHandCardFinish then
            self._onInsertHandCardFinish()
        end
    end

    local timeMove = 0.5
    -- 插入前先移动其他手牌，给要插的牌留出空位
    -- !!!重要：必须要保证这里的手牌移动要比移动要插入的手牌早结束，因为插入手牌移动结束后会刷新所有手牌
    if int_pos < index then
        -- 1. 如果空位在要插入位置的左边，从空位到插入位置左边的所有牌都要往左移一个单位
        for i=int_pos+1,index do
            local pos = self:getHandCardPosition(i-1)
            self._hand_cards[i]:stopAllActions()
            self._hand_cards[i]:runAction(CCMoveTo:create(0.15*timeMove,pos))
        end
    elseif int_pos > index then
        -- 2. 如果空位在要插入位置的右边，从插入位置到空位的所有牌都要往右移一个单位
        for i=index,int_pos-1 do
            local pos = self:getHandCardPosition(i+1)
            self._hand_cards[i]:stopAllActions()
            self._hand_cards[i]:runAction(CCMoveTo:create(0.15*timeMove,pos))
        end
    else
        -- 3. 如果插入位置和空位是一个位置，则其他牌不变
    end

    -- 移动最右边的牌
    if index == #other_cards+1 then
        -- 算出来还是移动到最后一张，直接移
        local pt = self:getHandCardPosition(index)
        --local x = pt.x
        --print(x)
        local move_array=CCArray:create()
        move_array:addObject(CCMoveTo:create(0.35*timeMove,pt))
        move_array:addObject(CCCallFunc:create(move_end_func))
        local move_action=CCSequence:create(move_array)
        card:stopAllActions()
        card:runAction(move_action)
    else
        -- 否则要从上方插入
        local pt = self:getHandCardPosition(index)
        local pos_x = pt.x
        local pos_y = pt.y
        local move_array=CCArray:create()
        move_array:addObject(CCMoveTo:create(0.15*timeMove,ccp(card:getPositionX(),pos_y+100)))
        move_array:addObject(CCMoveTo:create(0.075*timeMove*(#other_cards-index),ccp(pos_x,pos_y+100)))
        move_array:addObject(CCMoveTo:create(0.15*timeMove,ccp(pos_x,pos_y)))
        move_array:addObject(CCCallFunc:create(move_end_func))
        local move_action=CCSequence:create(move_array)
        card:stopAllActions()
        card:runAction(move_action)
    end
end

-- 添加一张牌到吃碰杠牌堆
---@param e_fromIndex UserIndex @用于显示箭头，如果是暗杠的，可以传nil，则不显示箭头
---@param e_kind HandleType
---@param byte_card int
---@param optional int_pos int
function CSelfCardsController:addHeapCard(e_fromIndex, e_kind, byte_card, int_pos)
    if not int_pos then
        int_pos = #self._heap_cards + 1
    end

    local heap = self:getHeapCard(int_pos)

    if heap then
        heap:setHandledData(e_fromIndex, e_kind, byte_card)
        if e_kind == HandleType.d_gang then
            -- 自己的暗杠，要根据配置来显示第四张牌
            if LayoutConfig.show_self_dark_gang_card_value then
                heap._handled_card[4]:setCardValue(byte_card)
            end
        end
        heap:showCards(self._corner_config, self._dark_card_value)
        -- 自己手牌的箭头需要根据是否有4张牌来调整，这里重新设箭头位置
        local c
        if e_kind == HandleType.d_gang or e_kind == HandleType.b_gang or e_kind == HandleType.a_gang then
            c = heap._handled_card[4]
        else
            c = heap._handled_card[2]
        end
        local pos = ccp(c:getPositionX(), c:getPositionY() + self._HEAP_ARROW_Y)
        local z = c:getZOrder()
        heap._arrow:setPosition(pos)
        heap._arrow:setScale(self._HEAP_ARROW_SCALE)
        heap._arrow:setZOrder(z+1)
    end
end

---@param int_pos int
function CSelfCardsController:removeHeapCard(int_pos)
    if self._heap_cards and #self._heap_cards > 0 then
        if not int_pos then
            int_pos = #self._heap_cards
        end
        self._heap_cards[int_pos]:hideCards()
    end
end

function CSelfCardsController:getHeapCard(int_pos)
    -- 先new一个新的
    if not self._heap_cards[int_pos] then
        local card_info,base_x,base_y,base_zorder = self:makeHeapCardLayoutInfo(int_pos)
        self._heap_cards[int_pos] = CHandledCards3D.new(card_info,base_x,base_y,base_zorder)
        for i = 1, 4 do
            self._view_root:addChild(self._heap_cards[int_pos]._handled_card[i])
        end
        self._view_root:addChild(self._heap_cards[int_pos])
        self._view_root:addChild(self._heap_cards[int_pos]._arrow)
    end
    return self._heap_cards[int_pos]
end

-- 添加一张牌到手牌，牌值是byte_card，位置是int_pos可省略，如果位置不是(最大index+1)，原来位置的牌会被替换
---@param byte_card int
---@param optional int_pos int
function CSelfCardsController:addHandCard(byte_card, int_pos)

    if not int_pos then
        int_pos = #self._hand_cards + 1
    end

    local hand_card = self:getHandCard(int_pos)

    if hand_card then
        local card_info,base_x,base_y,base_zorder = self:makeHandCardLayoutInfo(int_pos, self._lay)
        if self._lay_index ~= 0 and int_pos == self._lay_index then
            card_info,base_x,base_y,base_zorder = self:makeHandCardLayoutInfo(int_pos, true)
        end
        hand_card:init(card_info,base_x,base_y,base_zorder)
        hand_card:showCard(byte_card, self._corner_config[byte_card])
    end
end

---@param int_pos int
function CSelfCardsController:removeHandCard(int_pos)
    if self._hand_cards and #self._hand_cards > 0 then
        if not int_pos then
            int_pos = #self._hand_cards
        end

        self._hand_cards[int_pos]:hideCard()
    end
end

function CSelfCardsController:getHandCard(int_pos)
    -- 先new一个新的
    if not self._hand_cards[int_pos] then
        self._hand_cards[int_pos] = self:newHandCard()
        self._view_root:addChild(self._hand_cards[int_pos])
        self._hand_cards[int_pos]:setController(self)
        self._hand_cards[int_pos]:setCanTouch(true)
    end
    return self._hand_cards[int_pos]
end

---@param bool_can_touch bool @是否响应点击
---@param optional list_index table @{1,3,4,5,6...}哪些手牌会被设置，参数是序号组成的table，可以不传，不传则全部设置
function CSelfCardsController:setHandCardsCanTouch(bool_can_touch, list_index)
    if not self._hand_cards then return end
    if list_index then
        for i=1,#list_index do
            if self._hand_cards[list_index[i]] then
                self._hand_cards[list_index[i]]:setCanTouch(bool_can_touch)
            end
        end
    else
        for _,v in pairs(self._hand_cards) do
            if v then
                v:setCanTouch(bool_can_touch)
            end
        end
    end
end

---@param bool_can_touch bool @是否响应点击
---@param optional list_value table @{11,13,14,...}手牌牌值列表
function CSelfCardsController:setHandCardsCanTouchByValue(bool_can_touch, list_value)
    if not self._hand_cards then return end
    local list = self:getHandCardsIndexByValue(list_value)
    if list and #list>0 then
        self:setHandCardsCanTouch(bool_can_touch, list)
    end
end

---@param bool_can_out bool @是否可以被打出
---@param optional list_index table @{1,3,4,5,6...}哪些手牌会被设置，参数是序号组成的table，可以不传，不传则全部设置
function CSelfCardsController:setHandCardsCanOut(bool_can_out, list_index)
    if not self._hand_cards then return end
    if list_index then
        for i=1,#list_index do
            if self._hand_cards[list_index[i]] then
                self._hand_cards[list_index[i]]:setCanOut(bool_can_out)
            end
        end
    else
        for _,v in pairs(self._hand_cards) do
            if v then
                v:setCanOut(bool_can_out)
            end
        end
    end
end

---@param bool_can_out bool @是否可出
---@param optional list_value table @{11,13,14,...}手牌牌值列表
function CSelfCardsController:setHandCardsCanOutByValue(bool_can_out, list_value)
    if not self._hand_cards then return end
    local list = self:getHandCardsIndexByValue(list_value)
    if list and #list>0 then
        self:setHandCardsCanOut(bool_can_out, list)
    end
end

---@param bool_set_dark bool @是否变黑
---@param optional list_index table @{1,3,4,5,6...}哪些手牌会被设置，参数是序号组成的table，可以不传，不传则全部设置
function CSelfCardsController:setHandCardsDark(bool_set_dark, list_index)
    if not self._hand_cards then return end
    if list_index then
        for i=1,#list_index do
            if self._hand_cards[list_index[i]] then
                if bool_set_dark then
                    self._hand_cards[list_index[i]]:showMask()
                else
                    self._hand_cards[list_index[i]]:hideMask()
                end
            end
        end
    else
        for _,v in pairs(self._hand_cards) do
            if v then
                if bool_set_dark then
                    v:showMask()
                else
                    v:hideMask()
                end
            end
        end
    end
end

---@param bool_set_dark bool @是否变黑
---@param optional list_value table @{11,13,14,...}手牌牌值列表
function CSelfCardsController:setHandCardsDarkByValue(bool_set_dark, list_value)
    if not self._hand_cards then return end
    local list = self:getHandCardsIndexByValue(list_value)
    if list and #list>0 then
        self:setHandCardsDark(bool_set_dark, list)
    end
end

-- 通过手牌序号指定哪些牌显示听牌标记，未被指定的会隐藏，如果要全部隐藏可以传{}或nil
---@param optional list_index table @{1,3,4,5,6...}哪些手牌会被打上听牌标记，其余手牌则不显示标记，参数是序号组成的table，可以不传，不传则全部设置
function CSelfCardsController:setHandCardsTing(list_index)
    if not self._hand_cards then return end
    if list_index then
        self._ting_list = list_index
    else
        self._ting_list = {}
    end
    self:refreshHandCards()
end

-- 通过指定手牌值列表来指定哪些手牌需要显示听牌标记，未被指定的会隐藏，如果要全部隐藏可以传{}或nil
---@param optional list_value table @{11,13,14,...}手牌牌值如果在list里会被打上听牌标记
function CSelfCardsController:setHandCardsTingByValue(list_value)
    if not self._hand_cards then return end
    self._ting_list = self:getHandCardsIndexByValue(list_value)
    self:refreshHandCards()
end

function CSelfCardsController:setHandCardsTingDaByValue(list_value)
    if not self._hand_cards then return end
    self._ting_da_list = self:getHandCardsIndexByValue(list_value)
    self:refreshHandCards()
end

function CSelfCardsController:setHandCardsTingDuoByValue(list_value)
    if not self._hand_cards then return end
    self._ting_duo_list = self:getHandCardsIndexByValue(list_value)
    self:refreshHandCards()
end

---@return table @{1,2,4,5,7,8,...}
---@param list_value table @{11,13,14,...}
function CSelfCardsController:getHandCardsIndexByValue(list_value)
    local list = {}
    if list_value then
        for k,v in pairs(self._hand_cards) do
            if v then
                local found = false
                for i=1,#list_value do
                    if list_value[i] == v._card_value then
                        found = true
                        break
                    end
                end
                if found then
                    table.insert(list, k)
                end
            end
        end
    end
    return list
end

function CSelfCardsController:getHandCardPosition(int_pos)
    local card_info,base_x,base_y,base_zorder = self:makeHandCardLayoutInfo(int_pos, false)
    return ccp(base_x+card_info.ground.x, base_y+card_info.ground.y)
end

-- 不通过读取json，自己定义一个table来定制自己手牌
---@param int_pos int
---@param bool_lay bool @是否摊开
---@return card_info table
---@return base_x int
---@return base_y int
---@return base_zorder int
function CSelfCardsController:makeHandCardLayoutInfo(int_pos, bool_lay)
    local base_zorder = 1000
    local base_x = 0
    local base_y = 0

    local card_info = {}
    card_info.id = int_pos - 1
    card_info.ground = {}
    if bool_lay then
        card_info.ground.file = "bottom_extra_bg.png"
    else
        card_info.ground.file = "bottom_hand_1.png"
    end

    -- 如果有被选中的牌，被选中的牌会放大，右边的牌都会有一个位移
    local card_x_offset = 0
    local card_y_offset = 0
    local card_scale = self._HAND_CARD_SCALE
    if self._selected_index ~= 0 then
        local w = self._HAND_CARD_WIDTH
        local h = self._HAND_CARD_HEIGHT
        local s_c = self._SELECTED_HAND_CARD_SCALE
        local c = self._HAND_CARD_SCALE
        if int_pos == self._selected_index then
            card_x_offset = w * (s_c-c)/c * 0.5
            card_scale = self._SELECTED_HAND_CARD_SCALE
            card_y_offset = h * (s_c-c)/c * 0.5
        elseif int_pos > self._selected_index then
            card_x_offset = w * (s_c-c)/c
        end
    elseif type(self._selected_table) == type({}) and #self._selected_table > 0 then
        local w = self._HAND_CARD_WIDTH
        local h = self._HAND_CARD_HEIGHT
        local s_c = self._SELECTED_HAND_CARD_SCALE
        local c = self._HAND_CARD_SCALE
        local index = 0;
        local isFind = false;
        for k,v in pairs(self._selected_table) do
            if int_pos == v then
                isFind = true;
            end
            if int_pos >= v then
                index = k;
            else
                break;
            end
        end
        if index > 0 then
            if isFind then
                card_y_offset = h * (s_c-c)/c
            end
        end
    end
    -- 手牌X坐标，通过计算自己是第几张决定
    card_info.ground.x = self._hand_cards_start_x + (int_pos-1)*(self._HAND_CARD_WIDTH-self._HAND_CARD_EDGE) + card_x_offset

    local cards_count = #cards_3d_logic:getCards(UserIndex.down, LayoutCardsType.hand)
    if int_pos == cards_count and cards_count % 3 == 2 then
        -- 是最后一张且%3==2
        card_info.ground.x = card_info.ground.x + self._CATCH_HAND_CARD_SPACE
    end
    card_info.ground.y = self._HAND_CARD_BASE_Y + card_y_offset
    card_info.ground.scale = card_scale

    card_info.card = {}
    card_info.card.file = "hand_card_{NUM}.png"
    card_info.card.x = -36
    if bool_lay then
        card_info.card.y = -38
    else
        card_info.card.y = -55
    end
    card_info.card.scale = 0.7

    card_info.corner = {}
    card_info.corner.file = "corner_{NUM}_7.png"
    card_info.corner.x = -21
    if bool_lay then
        card_info.corner.y = -16
    else
        card_info.corner.y = -28
    end
    card_info.corner.scale = 1

    card_info.zorder = 16

    return card_info, base_x, base_y, base_zorder
end

-- 不通过读取json，自己定义一个table来定制自己吃碰杠牌堆
---@param int_pos
---@return card_info table
---@return base_x int
---@return base_y int
---@return base_zorder int
function CSelfCardsController:makeHeapCardLayoutInfo(int_pos)
    local base_zorder = 1000
    local base_x = 0
    local base_y = 0

    local card_info = {}

    card_info.id = int_pos - 1

    card_info.direct = {}
    card_info.direct.file = "direct_bottom_{NUM}.png"
    card_info.direct.x = 0  -- 自己手牌的箭头需要根据是否有4张牌来调整，这里随便给个值
    card_info.direct.y = 0
    card_info.direct.scale = 1
    card_info.direct.zorder = 4

    card_info.items = {}
    local heap_count = #cards_3d_logic:getCards(UserIndex.down, LayoutCardsType.heap)
    for i=1,4 do
        card_info.items[i] = self:makeSingleHeapCardLayoutInfo(heap_count, int_pos, i)
    end

    return card_info, base_x, base_y, base_zorder
end

---@return single_card table @{"id":0,
---"ground":{"file":"bottom_extra_bg.png","back":"bottom_extra_an.png","x":-414,"y":-5,"scale":0.8},
---"card":{"file":"hand_card_{NUM}.png","x":-36,"y":-37,"scale":0.7},
---"corner":{"file":"corner_{NUM}_7.png","x":-21,"y":-16,"scale":1.00},
---"zorder":0}
---@param int_heap_count @总共有N个heap --暂时用不到
---@param int_heap_index @第N个heap
---@param int_index int @这个heap中的第N张牌
function CSelfCardsController:makeSingleHeapCardLayoutInfo(int_heap_count, int_heap_index, int_index)
    local single_card = {}

    single_card.id = int_index-1

    single_card.ground = {}
    single_card.ground.file = "bottom_extra_bg.png"
    single_card.ground.back = "bottom_extra_an.png"
    single_card.ground.x = self._heap_pos[int_heap_index].x + self._heap_card_pos[int_index].x
    single_card.ground.y = self._heap_pos[int_heap_index].y + self._heap_card_pos[int_index].y
    single_card.ground.scale = self._HEAP_CARD_SCALE

    single_card.card = {}
    single_card.card.file = "hand_card_{NUM}.png"
    single_card.card.x = -36
    single_card.card.y = -37
    single_card.card.scale = 0.7

    single_card.corner = {}
    single_card.corner.file = "corner_{NUM}_7.png"
    single_card.corner.x = -21
    single_card.corner.y = -16
    single_card.corner.scale = 0.7

    single_card.zorder = (int_heap_index-1) * 4 + (int_index-1)
    return single_card
end

-- 统一处理某张牌被点击、拖动消息
---@param sender table
---@param touchType int @_G.TOUCH_EVENT_BEGAN|_G.TOUCH_EVENT_MOVED|_G.TOUCH_EVENT_ENDE|_G.TOUCH_EVENT_CANCELED
---@param card CSelfHandCard3D @收到输入消息的牌
function CSelfCardsController:touchEvent(sender,touchType,card)
    -- 有任何输入，都要通知外部，处理比如关闭菜单等
    if self._onAnyTouchEventFunc then self._onAnyTouchEventFunc(sender, touchType) end

    if self._select_type == 1 then
        if touchType~=_G.TOUCH_EVENT_ENDED then return; end
        local index = self:FindCardIndex(card)
        local isFind = false;
        local findIndex = 0;
        for k,v in pairs(self._selected_table) do
            if v <= index then 
                findIndex = k;
            end
            if v == index then
                isFind = true;
            end
            if v >= index then 
                break;
            end
        end
        
        if isFind then
            -- 之前已经点击过这张牌了
            table.remove(self._selected_table,findIndex);
            self:refreshHandCards()
        else
            print("findIndex:"..findIndex);
            -- 之前没有点击过，重新设选中index
            table.insert(self._selected_table,findIndex + 1,index);
            self:refreshHandCards()
        end
        return;
    end

    self:setSelectedTable({});
    if touchType==_G.TOUCH_EVENT_BEGAN then
        self._touch_start_pos = card._touch_pos
    elseif touchType==_G.TOUCH_EVENT_MOVED then
        if self._drag_began == false then
            if self._touch_start_pos == nil then return;end
            local diff = math.abs(card._touch_pos.x-self._touch_start_pos.x)+math.abs(card._touch_pos.y-self._touch_start_pos.y)
            if diff > 20 then
                -- 认为开始拖动牌，临时把zorder设高，结束后刷新会重置
                self._drag_began = true

                -- 把放大的牌恢复正常大小
                self:setSelectedIndex(0)
                -- 刷新以后再设zorder，否则会被覆盖掉
                card:setZOrder(2000)
                -- 被拖动的牌不显示听牌标记
                card:hideTing()
            end
        end
        if self._drag_began then
            -- 更新位置
            card:setPosition(card._touch_pos)
        end
    elseif touchType==_G.TOUCH_EVENT_ENDED then
        if self._drag_began == false then
            -- 没有开始拖动，认为是点击了这张牌
            self:OnCardClicked(card)
            card:hideTing()
        else
            -- 拖动之后的END，判断手牌位置来决定是否是出牌操作
            self._touch_start_pos = nil
            self._drag_began = false
            if card:getPositionY() > 200 and card._bool_can_out then
                -- 出牌，不用刷新手牌，出牌有插牌动画，动画结束后会刷新
                if self._onMoveCardOut then
                    self._onMoveCardOut(card)
                end
            else
                -- 刷新手牌
                self:refreshHandCards()
            end
        end
    elseif touchType==_G.TOUCH_EVENT_CANCELED then
        self._touch_start_pos = nil
        self._drag_began = false
        -- 回到手牌
        self:refreshHandCards()
    end
end

function CSelfCardsController:OnCardClicked(card)
    local index = self:FindCardIndex(card)
    if self._selected_index == index and card._bool_can_out then
        -- 之前已经点击过这张牌了，又被点击，则出牌
        self:setSelectedIndex(0)
        self:refreshHandCards()
        if self._onClickCardOut then
            self._onClickCardOut(card)
        end
    else
        -- 之前没有点击过，重新设选中index
        self:setSelectedIndex(index)
        self:refreshHandCards()
        if self._onClickCard then
            self._onClickCard(card)
        end
    end
end

-- 通过一张card来找到对应的index
function CSelfCardsController:FindCardIndex(card)
    for i=1,#self._hand_cards do
        if self._hand_cards[i] == card then
            return i
        end
    end
    return 0
end

function CSelfCardsController:setSelectedIndexFunc(func_onSelectedIndex)
    if func_onSelectedIndex and type(func_onSelectedIndex) == "function" then
        self._onSelectedIndexFunc=func_onSelectedIndex
    end
end

function CSelfCardsController:setOnAnyTouchEventFunc(func_onAnyTouchEvent)
    if func_onAnyTouchEvent and type(func_onAnyTouchEvent) == "function" then
        self._onAnyTouchEventFunc=func_onAnyTouchEvent
    end
end

-- 把手牌拖到上方出牌的callback
---@param func_onMoveCardOut function @Sample:
---CSelfCardsController:setOnMoveCardOutFunc(function(card)
---    local card_value = card:getCardValue()
---    card:setCanTouch(false)
---    ...
---end)
function CSelfCardsController:setOnMoveCardOutFunc(func_onMoveCardOut)
    if func_onMoveCardOut and type(func_onMoveCardOut) == "function" then
        self._onMoveCardOut=func_onMoveCardOut
    end
end

-- 两次点击手牌出牌的callback
---@param func_onClickCardOut function @Sample:
---CSelfCardsController:setOnClickCardOutFunc(function(card)
---    local card_value = card:getCardValue()
---    card:setCanTouch(false)
---    ...
---end)
function CSelfCardsController:setOnClickCardOutFunc(func_onClickCardOut)
    if func_onClickCardOut and type(func_onClickCardOut) == "function" then
        self._onClickCardOut=func_onClickCardOut
    end
end

-- 一次点击手牌，手牌选中变大的callback
---@param func_onClickCard function @Sample:
---CSelfCardsController:setOnClickCardFunc(function(card)
---    local card_value = card:getCardValue()
---    ...
---end)
function CSelfCardsController:setOnClickCardFunc(func_onClickCard)
    if func_onClickCard and type(func_onClickCard) == "function" then
        self._onClickCard=func_onClickCard
    end
end

---@param func_findCardInHandPosition function @Sample:
---CSelfCardsController:setFindCardInHandPositionFunc(function(cards,value)
---    local current_hand_cards = cards -- {11,12,13,14...}
---    local card_value = value -- 11..19|21..29|...
---    ...
---end)
function CSelfCardsController:setFindCardInHandPositionFunc(func_findCardInHandPosition)
    if func_findCardInHandPosition and type(func_findCardInHandPosition) == "function" then
        self._findCardInHandPosition=func_findCardInHandPosition
    end
end

function CSelfCardsController:setOnInsertHandCardFinishFunc(func_onInsertHandCardFinish)
    if func_onInsertHandCardFinish and type(func_onInsertHandCardFinish) == "function" then
        self._onInsertHandCardFinish=func_onInsertHandCardFinish
    end
end

return CSelfCardsController











































