require ("src/cards_3d_config")
---@class CSelfHandCard3D @自己的单张手牌，可以设牌墙（竖立正面），可以设摊开
CSelfHandCard3D=class("CSelfHandCard3D", function() return ImageView:create() end)
function CSelfHandCard3D:ctor(...)
    self._controller=nil
    self._size={}
    self._scale=1
    self._card_value=InvalidValue
    self._visual_value=InvalidValue
    self._message_value=InvalidValue
    self._type=HandleType.null
    self._card_img=nil
    self._ground_path = nil
    self._touch_paused = false
    self._back_path = "bottom_extra_an.png"
    self._is_dark = false
    if #{...} == 0 then
        self:init(self:makeDummyHandCardLayoutInfo())
    else
        self:init(...)
    end
end

-- 要可以被多次初始化
---@param card_info table
---@param base_x int
---@param base_y int
---@param base_zorder int
function CSelfHandCard3D:init(card_info,base_x,base_y,base_zorder)
    if not self._card_img then
        self._card_img=control_tools.newImg({})
        self:addChild(self._card_img)
    end

    if not self._corner_img then
        self._corner_img=control_tools.newImg({})
        self:addChild(self._corner_img)
    end

    if not self._card_mask then
        self._card_mask=control_tools.newImg({})
        self:addChild(self._card_mask)
    end

    if not self._ting then
        self._ting=control_tools.newImg({path="ting_tip.png", ctype=1})

        self:addChild(self._ting)
    end

    if card_info ~= nil then
        -- 底
        self._ground_path = card_info.ground.file
        self:loadTexture(self._ground_path,1)
        self:setPosition(ccp(base_x+card_info.ground.x, base_y+card_info.ground.y))
        self:setScale(card_info.ground.scale)
        self:setZOrder(base_zorder+card_info.zorder)

        -- 需要考虑没有牌面的情况
        if card_info.card ~= nil then
            -- 牌面
            local card_x = self:getSize().width / 2 + card_info.card.x
            local card_y = self:getSize().height / 2 + card_info.card.y
            self._card_img:setPosition(ccp(card_x, card_y))
            self._card_img:setScale(card_info.card.scale)
            self._card_path=card_info.card.file
        end

        -- 角标
        if card_info.corner then
            self.corner_x = self:getSize().width / 2 + card_info.corner.x
            self.corner_y = self:getSize().height / 2 + card_info.corner.y
            self.corner_scale = card_info.corner.scale
            self._corner_img:setPosition(ccp(self.corner_x, self.corner_y))
            self._corner_img:setScale(card_info.corner.scale)
            self._corner_path=card_info.corner.file
        end

        -- mask
        self._card_mask:loadTexture(self._ground_path,1)
        self._card_mask:setPosition(ccp(0, 0))
        self._card_mask:setScale(1)
        self._card_mask:setOpacity(100)
        self._card_mask:setZOrder(base_zorder+card_info.zorder)
        self._card_mask:setVisible(false)
        self._card_mask:runAction(CCTintTo:create(0,20,20,20))

        -- ting
        self._ting:setPosition(ccp(0,self:getSize().height*0.5+self._ting:getSize().height*0.5))
        self._ting:setZOrder(base_zorder+card_info.zorder)
        self._ting:setVisible(false)
    end

    if self._is_dark then
        self:showMask()
    else
        self:hideMask()
    end

    self:setVisible(false)
end

-- 不通过读取json，自己定义一个table来定制自己手牌
---@param int_pos
---@return card_info table
---@return base_x int
---@return base_y int
---@return base_zorder int
function CSelfHandCard3D:makeDummyHandCardLayoutInfo()
    local base_zorder = 300
    local base_x = 0
    local base_y = 0

    local card_info = {}
    card_info.id = 0

    card_info.ground = {}
    card_info.ground.file = "bottom_extra_bg.png"
    card_info.ground.x = 0
    card_info.ground.y = 0
    card_info.ground.scale = 1.0

    card_info.card = {}
    card_info.card.file = "hand_card_{NUM}.png"
    card_info.card.x = -36
    card_info.card.y = -38
    card_info.card.scale = 0.7

    card_info.corner = {}
    card_info.corner.file = "corner_{NUM}_7.png"
    card_info.corner.x = -21
    card_info.corner.y = -16
    card_info.corner.scale = 1

    card_info.zorder = 16

    return card_info, base_x, base_y, base_zorder
end

function CSelfHandCard3D:setController(controller)
    self._controller = controller
end

---@param param_value byte @牌值
---@param param_corner_id int @角标，没有为0或nil，财神为1，鸡为2，后续待添加
---@param param_show_mask bool @是否显示mask
function CSelfHandCard3D:setCardValue(param_value,param_corner_id,param_show_mask)
    self._card_value=param_value
    if param_value == InvalidValue then
        self._visual_value = InvalidValue
        self._card_img:setVisible(false)
        self:hideCorner()
        self:hideMask()
    else
        if CardExchangeTable[param_value] ~= nil then
            param_value = CardExchangeTable[param_value]
        end
        self._visual_value = param_value
        if self._card_path then
            if (self._visual_value >= 10 and self._visual_value <= 47) or (self._visual_value >= 50 and self._visual_value <= 58) then
                local img_path = string.gsub(self._card_path, "{NUM}", self._visual_value)
                self._card_img:loadTexture(img_path,1)
                self._card_img:setVisible(true)
                self._card_img:setOpacity(255)
            end
        end

        self:setOpacity(255)
        if self._corner_img then
            self._corner_img:setOpacity(255)
        end

        if param_corner_id and param_corner_id ~= 0 then
            self:showCorner(param_corner_id)
        else
            self:hideCorner()
        end

        -- 如果没有param_show_mask参数，则不做变化
        if param_show_mask then
            if param_show_mask==true then
                self:showMask()
            else
                self:hideMask()
            end
        else
            if self._is_dark == true then
                self:showMask()
            else
                self:hideMask()
            end
        end
    end
end

function CSelfHandCard3D:getCardValue()
    return self._card_value
end

function CSelfHandCard3D:setCanTouch(bool_touch)
    if bool_touch==true then
        self:setTouchEnabled(true)
        self:addTouchEventListener(function(sender,touchType) self:touchEvent(sender,touchType) end)
    else
        self:setTouchEnabled(false)
    end
end

function CSelfHandCard3D:setCanOut(bool_can_out)
    self._bool_can_out = bool_can_out
end

function CSelfHandCard3D:showTing(index)
    if self._ting then
        if index == 1 then
            self._ting:loadTexture("ting_tip.png",1);
        elseif index == 2 then
            self._ting:loadTexture("ting_tip_da.png",1);
        elseif index == 3 then
            self._ting:loadTexture("ting_tip_duo.png",1);
        end
        self._ting:setVisible(true)
    end
end

function CSelfHandCard3D:hideTing()
    if self._ting then
        self._ting:setVisible(false)
    end
end

function CSelfHandCard3D:setMessageValue(param_value)
    self._message_value=param_value
end

function CSelfHandCard3D:setType(param_type)
    self._type=param_type
end

function CSelfHandCard3D:touchEvent(sender,touchType)
    if self._touch_paused then return end

    if touchType==_G.TOUCH_EVENT_BEGAN then
        self._touch_pos = self:getTouchStartPos()
    elseif touchType==_G.TOUCH_EVENT_MOVED then
        self._touch_pos = self:getTouchMovePos()
    elseif touchType==_G.TOUCH_EVENT_ENDED then
        self._touch_pos = self:getTouchEndPos()
    elseif touchType==_G.TOUCH_EVENT_CANCELED then
        self._touch_pos = self:getTouchMovePos()
    else
        self._touch_pos = nil
    end
    if self._controller then
        self._controller:touchEvent(sender, touchType, self)
    end
end

function CSelfHandCard3D:showMask()
    self._is_dark = true
    self._card_mask:setOpacity(100);
    self._card_mask:setVisible(true)
end

function CSelfHandCard3D:hideMask()
    self._is_dark = false
    self._card_mask:setVisible(false)
end

-- 设置角标
---@param corner_id int @1代表财神，2代表鸡，3-6代表1-4数字角标
function CSelfHandCard3D:showCorner(corner_id)
    if corner_id >= 1 and corner_id <= 7 then
        local img_path = string.gsub(self._corner_path, "{NUM}", corner_id)
        self._corner_img:loadTexture(img_path,1)

        self._corner_img:setVisible(true)
        self._corner_img:setPosition(ccp(self.corner_x, self.corner_y))
        self._corner_img:setScale(self.corner_scale)
    end
end

-- 隐藏角标
function CSelfHandCard3D:hideCorner()
    self._corner_img:setVisible(false)
end

function CSelfHandCard3D:getPosX()
    return self._posx
end

function CSelfHandCard3D:getPosY()
    return self._posy
end

-- 设置牌面透明度
---@param int_opacity int @透明度
function CSelfHandCard3D:setCardOpacity(int_opacity)
    self:setOpacity(int_opacity)
    self._card_img:setOpacity(int_opacity)
    self._corner_img:setOpacity(int_opacity)
end

-- 渐显
---@param float_time float @渐显时间
function CSelfHandCard3D:fadeIn(float_time)
    self:stopAllActions()
    self:runAction(CCFadeIn:create(float_time))
    self._card_img:stopAllActions()
    self._card_img:runAction(CCFadeIn:create(float_time))
    self._corner_img:stopAllActions()
    self._corner_img:runAction(CCFadeIn:create(float_time))
end

-- 渐隐
---@param float_time float @渐隐时间
function CSelfHandCard3D:fadeOut(float_time)
    self:stopAllActions()
    self:runAction(CCFadeOut:create(float_time))
    self._card_img:stopAllActions()
    self._card_img:runAction(CCFadeOut:create(float_time))
    self._corner_img:stopAllActions()
    self._corner_img:runAction(CCFadeOut:create(float_time))
end

-- 显示牌背
function CSelfHandCard3D:showBack()
    self:setCardValue(InvalidValue)
    self:hideCorner()
    self:hideMask()
    self:hideTing()
    self:loadTexture(self._back_path, 1)
end

-- 设置牌值并显示
---@param param_value byte @牌值
---@param param_corner_id int @角标，没有为0或nil，财神为1，后续待添加
---@param param_show_mask bool @是否显示mask
function CSelfHandCard3D:showCard(param_value,param_corner_id,param_show_mask)
    if param_value == nil then
        param_value = self._card_value
    end
    self:setCardValue(param_value,param_corner_id,param_show_mask)
    self:setVisible(true)
end

-- 隐藏
function CSelfHandCard3D:hideCard()
    self._card_value=InvalidValue
    self._visual_value=InvalidValue
    self:setVisible(false)
end

function CSelfHandCard3D:destroy()

end

return CSelfHandCard3D
