require ("src/cards_3d_config")
---@class CTiledCard3D @单张牌（不包括自己吃碰杠堆牌），可以设正面显示或背面显示，用在出牌堆，花牌堆，吃碰杠牌堆，出牌动画
CTiledCard3D=class("CTiledCard3D", function() return ImageView:create() end)
function CTiledCard3D:ctor(...)
	self._size={}
	self._card_value=InvalidValue
    self._visual_value=InvalidValue
    self.card_x = nil
    self.card_y = nil
    self.card_scale = nil
    -- 牌面的路径（带{NUM{}）
    self._card_path = nil
    -- 牌底的路径
    self._back_up_path = nil
    -- 牌背面朝上的路径
    self._back_down_path = nil
	self:init(...)
end

-- 根据配置初始化一个card
---@param card_info table
---@param base_x int
---@param base_y int
---@param base_zorder int
function CTiledCard3D:init(card_info,base_x,base_y,base_zorder)
    if not card_info then return end

    if false then
        print("CTiledCard3D base_x "..base_x)
        print("CTiledCard3D base_y "..base_y)
        print("CTiledCard3D base_zorder "..base_zorder)
        print("CTiledCard3D card_info.id "..card_info.id)
        print("CTiledCard3D card_info.zorder "..card_info.zorder)
        print("CTiledCard3D card_info.card.file "..card_info.card.file)
        print("CTiledCard3D card_info.card.x "..card_info.card.x)
        print("CTiledCard3D card_info.card.y "..card_info.card.y)
        print("CTiledCard3D card_info.card.scale "..card_info.card.scale)
        print("CTiledCard3D card_info.ground.file "..card_info.ground.file)
        print("CTiledCard3D card_info.ground.x "..card_info.ground.x)
        print("CTiledCard3D card_info.ground.y "..card_info.ground.y)
        print("CTiledCard3D card_info.ground.scale "..card_info.ground.scale)
    end

    if card_info.ground.back then
        self._back_down_path = card_info.ground.back
    end

    -- 底
    self._ground_path = card_info.ground.file
    self:loadTexture(self._ground_path,1)
    self.back_x = base_x+card_info.ground.x
    self.back_y = base_y+card_info.ground.y
    self.back_scale = card_info.ground.scale
    self:setPosition(ccp(self.back_x, self.back_y))
    self:setScale(self.back_scale)
    self:setZOrder(base_zorder+card_info.zorder)


    -- 牌面
    self.card_x = self:getSize().width / 2 + card_info.card.x
    self.card_y = self:getSize().height / 2 + card_info.card.y
    self.card_scale = card_info.card.scale
    self._card_img=control_tools.newImg({})
    self._card_img:setPosition(ccp(self.card_x, self.card_y))
    self._card_img:setScale(card_info.card.scale)
    self:addChild(self._card_img)
    self._card_path=card_info.card.file

    -- 角标
    self.corner_x = self:getSize().width / 2 + card_info.corner.x
    self.corner_y = self:getSize().height / 2 + card_info.corner.y
    self.corner_scale = card_info.corner.scale
    self._corner_img=control_tools.newImg({})
    self._corner_img:setPosition(ccp(self.corner_x, self.corner_y))
    self._corner_img:setScale(card_info.corner.scale)
    self:addChild(self._corner_img)
    self._corner_path=card_info.corner.file

    -- mask
    self._card_mask=ImageView:create()
    self._card_mask:loadTexture(self._ground_path,1)
    self._card_mask:setPosition(ccp(0, 0))
    self._card_mask:setScale(1)
    self._card_mask:setOpacity(100)
    self._card_mask:setZOrder(base_zorder+card_info.zorder)
    self._card_mask:setVisible(false)
    self._card_mask:runAction(CCTintTo:create(0,20,20,20))
    self:addChild(self._card_mask)

    self:setVisible(false)
end

---@param param_value byte @牌值
---@param param_corner_id int @角标，没有为0或nil，财神为1，后续待添加
---@param param_show_mask bool @是否显示mask
function CTiledCard3D:setCardValue(param_value,param_corner_id,param_show_mask)
    self._card_value=param_value
    if param_value == InvalidValue then
        self._card_img:setVisible(false)
    else
        if CardExchangeTable[param_value] ~= nil then
            param_value = CardExchangeTable[param_value]
        end
        self._visual_value = param_value
        if (self._visual_value >= 10 and self._visual_value <= 47) or (self._visual_value >= 50 and self._visual_value <= 58) then
            local img_path = string.gsub(self._card_path, "{NUM}", self._visual_value)
            self._card_img:loadTexture(img_path,1)
        end

        self:loadTexture(self._ground_path,1)
        self._card_img:setVisible(true)
        self._card_img:setPosition(ccp(self.card_x, self.card_y))
        self._card_img:setScale(self.card_scale)

        if param_corner_id and param_corner_id ~= 0 then
            self:showCorner(param_corner_id)
        else
            self:hideCorner()
        end

        if param_show_mask and param_show_mask==true then
            self:showMask()
        else
            self:hideMask()
        end
    end
end

function CTiledCard3D:getCardValue()
    if self._card_value then
        return self._card_value
    else
        return InvalidValue
    end
end

function CTiledCard3D:getCardVisualValue()
    if self._visual_value then
        return self._visual_value
    else
        return InvalidValue
    end
end

---@param param_value byte @牌值
---@param param_corner_id int @角标，没有为0或nil，财神为1，后续待添加
---@param param_show_mask bool @是否显示mask
function CTiledCard3D:showCard(param_value,param_corner_id,param_show_mask)
    if param_value == nil then
        param_value = self:getCardValue()
    end
    self:setCardValue(param_value,param_corner_id,param_show_mask)
    self:setVisible(true)
end

function CTiledCard3D:hideCard()
    self:setVisible(false)
    self._card_value=InvalidValue
    self._visual_value=InvalidValue
    self:hideMask()
    self:hideCorner()
end

-- 显示背面向上
function CTiledCard3D:showBack()
    if self._back_down_path then
        self:loadTexture(self._back_down_path,1)
        self:setScale(self.back_scale)
        self:setCardValue(InvalidValue)
        self._visual_value=InvalidValue
    end
end

function CTiledCard3D:showMask()
    self._card_mask:setVisible(true)
end

function CTiledCard3D:hideMask()
    self._card_mask:setVisible(false)
end

function CTiledCard3D:showCorner(param_corner_id)
    if param_corner_id >= 1 and param_corner_id <= 7 then
        local img_path = string.gsub(self._corner_path, "{NUM}", param_corner_id)
        self._corner_img:loadTexture(img_path,1)

        self._corner_img:setVisible(true)
        self._corner_img:setPosition(ccp(self.corner_x, self.corner_y))
        self._corner_img:setScale(self.corner_scale)
    end
end

function CTiledCard3D:hideCorner()
    self._corner_img:setVisible(false)
end

function CTiledCard3D:destroy()

end

return CTiledCard3D
