require ("src/cards_3d_config")
require ("src/g_tools/control_tools")
require ("src/g_tools/game_tools")
require ("src/g_tools/files")
local CTiledCard3D=require("src/Controls/3D/CTiledCard3D")
local CHandCard3D=require("src/Controls/3D/CHandCard3D")
local CHandledCards3D=require("src/Controls/3D/CHandledCards3D")
local CSelfCardsController=require("src/Controls/3D/CSelfCardsController")
local CHandleSelect3D=require("src/Controls/3D/CHandleSelect3D")
local CCardPoint3D=require("src/Controls/3D/CCardPoint3D")
local cards_3d_logic=require("src/cards_3d_logic")
---@class cards_3d_layout
cards_3d_layout=class("cards_3d_layout")

-- 公共接口
function cards_3d_layout:ctor(...)
    cards_3d_layout:init(...)
end

-- 初始化，进游戏先调一次
---@param view_root ImageView @游戏创建的ImageView，用作父层级
---@param view_size table @{width=1000,height=750}
---@param table_3d_layout table_3d_layout @框架提供的table_3d_layout，必须已经初始化过
---@param int_user_count int @游戏的用户数，取2,3,4
---@param int_max_handcard_count int @游戏的最大手牌数，取14/17
function cards_3d_layout:init(view_root, view_size, table_3d_layout, int_user_count, int_max_handcard_count)
    self:insertLog("init", nil, nil, nil, int_user_count, int_max_handcard_count)

    -- 默认人数4，手牌数14
    if not int_user_count then int_user_count = 4 end
    if not int_max_handcard_count then int_max_handcard_count = 14 end
    -- 3人使用4人配置
    if int_user_count == 3 then int_user_count = 4 end

    CCSpriteFrameCache:sharedSpriteFrameCache():addSpriteFramesWithFile(Layout3DpListPath.."cards.plist", Layout3DpListPath.."cards.png")
    CCSpriteFrameCache:sharedSpriteFrameCache():addSpriteFramesWithFile(Layout3DpListPath.."backs.plist", Layout3DpListPath.."backs.png")
    CCSpriteFrameCache:sharedSpriteFrameCache():addSpriteFramesWithFile(Layout3DpListPath.."corners.plist", Layout3DpListPath.."corners.png")

    CCArmatureDataManager:sharedArmatureDataManager():removeArmatureFileInfo(Layout3DAnimationPath.."dh_cpght.ExportJson")
    CCArmatureDataManager:sharedArmatureDataManager():addArmatureFileInfo(Layout3DAnimationPath.."dh_cpght.ExportJson")

    self._view_root = view_root
    self._view_size = view_size
    self._table_3d_layout = table_3d_layout
    self._layout = {}
    self._corner_config = {}
    self._dark_card_value = nil
    self._hand_cards = {{},{},{},{}}    -- 手牌ImageView资源数组，一旦加载就不释放，直到游戏退出到大厅
    self._heap_cards = {{},{},{},{}}    -- 吃碰杠牌堆ImageView资源数组，一旦加载就不释放，直到游戏退出到大厅
    self._out_cards = {{},{},{},{}}     -- 出牌堆ImageView资源数组，一旦加载就不释放，直到游戏退出到大厅
    self._hua_cards = {{},{},{},{}}     -- 花牌ImageView资源数组，一旦加载就不释放，直到游戏退出到大厅
    self._tmp_moving_cards = {}
    self._handle_ani = {}
    self._card_mask = {}
    for i=1,int_user_count do
        table.insert(self._card_mask, false);
    end

    local path = CCFileUtils:sharedFileUtils():fullPathForFilename(Layout3DPath)
    path = path.."layout_"..int_user_count.."_"..int_max_handcard_count.."/"
    self:initLayout(path, view_size)

    CSelfCardsController:init(view_root, view_size, int_max_handcard_count, self._hand_cards[UserIndex.down], self._heap_cards[UserIndex.down])

    cards_3d_logic:init()

    -- 注册默认callback
    -- 有任意框架输入
    CSelfCardsController:setOnAnyTouchEventFunc(function(sender,event)
        if self._on_any_touch_event_func then
            self._on_any_touch_event_func(sender,event)
        end
    end)

    CSelfCardsController:setSelectedIndexFunc(function(int_index)
        local handcards = cards_3d_logic:getCards(UserIndex.down, LayoutCardsType.hand)
        if handcards and #handcards >= int_index then
            self:setDarkCardValue(handcards[int_index])
        end
    end)

    -- 当把一张手牌移到屏幕中间放开，出牌时调用
    CSelfCardsController:setOnMoveCardOutFunc(function(card)
        local byte_card = card:getCardValue()
        self:outCard(UserIndex.down, byte_card, card._touch_pos, CSelfCardsController:FindCardIndex(card))
        if self._move_card_out_func then
            self._move_card_out_func(byte_card)
        end
    end)

    -- 当点击一张手牌两次，出牌时调用
    CSelfCardsController:setOnClickCardOutFunc(function(card)
        local byte_card = card:getCardValue()
        self:outCard(UserIndex.down, byte_card, card._touch_pos, CSelfCardsController:FindCardIndex(card))
        -- 在上一步card._card_value被改写了，所以先保存再使用
        if self._click_card_out_func then
            self._click_card_out_func(byte_card)
        end
    end)

    -- 当点击一张手牌，选中变大时调用
    CSelfCardsController:setOnClickCardFunc(function(card)
        local byte_card = card:getCardValue()
        if self._click_card_func then
            self._click_card_func(byte_card)
        end
    end)

    -- 当摸上来的手牌在出牌后移动到合适位置，界面刷新之前调用
    CSelfCardsController:setOnInsertHandCardFinishFunc(function()

    end)

    cards_3d_logic:setSortCardsFunc(function(list_card)
        if self._sort_cards_func then
            return self._sort_cards_func(list_card)
        else
            return nil
        end
    end)

    self._handle_select = CHandleSelect3D.new(view_size)
    self._handle_select:setPosition(ccp(0,0))
    self._handle_select:setZOrder(LayoutZOrder.HandleSelect)
    view_root:addChild(self._handle_select)
    self._handle_select:setOnResultFunc(function(arg)
        if self._handle_clicked_func then
            self._handle_clicked_func(arg)
        end
    end)
end

-- 清空所有牌并重置状态，每盘游戏开局调一次！
function cards_3d_layout:reset()
    self:insertLog("reset")
    for i=1,4 do
        for j, v in pairs(self._hand_cards[i]) do
            v:hideCard()
        end
        for j, v in pairs(self._heap_cards[i]) do
            v:hideCards()
        end
        for j, v in pairs(self._out_cards[i]) do
            v:hideCard()
        end
        for j, v in pairs(self._hua_cards[i]) do
            v:hideCard()
        end
    end

    cards_3d_logic:reset()
    CSelfCardsController:reset()

    self._corner_config = {}
    if self._last_out_card_indicator then
        self._last_out_card_indicator:hidePoint()
    end
end

-- 释放资源，游戏退出时调用
function cards_3d_layout:destroy()
    self:insertLog("destroy")
    for i=1,4 do
        for j, v in pairs(self._hand_cards[i]) do
            v:destroy()
        end
        self._hand_cards[i] = {}
        for j, v in pairs(self._heap_cards[i]) do
            v:destroy()
        end
        self._heap_cards[i] = {}
        for j, v in pairs(self._out_cards[i]) do
            v:destroy()
        end
        self._out_cards[i] = {}
        for j, v in pairs(self._hua_cards[i]) do
            v:destroy()
        end
        self._hua_cards[i] = {}
    end
    self._tmp_moving_cards = {}

    if self._time_schedule then
        CCDirector:sharedDirector():getScheduler():unscheduleScriptEntry(self._time_schedule)
        self._time_schedule=nil
    end

    cards_3d_logic:destroy()

    CCSpriteFrameCache:sharedSpriteFrameCache():removeSpriteFramesFromFile(Layout3DpListPath.."cards.plist")
    CCSpriteFrameCache:sharedSpriteFrameCache():removeSpriteFramesFromFile(Layout3DpListPath.."backs.plist")
    CCSpriteFrameCache:sharedSpriteFrameCache():removeSpriteFramesFromFile(Layout3DpListPath.."corners.plist")

    CCArmatureDataManager:sharedArmatureDataManager():removeArmatureFileInfo(Layout3DAnimationPath.."dh_cpght.ExportJson")
end

-- 打印所有调用log
function cards_3d_layout:printLog()
    if LayoutConfig.log_enable and self._log then
        print(self._log)
    end
end

-- 一次性设置手牌，要求已排序
---@param e_index UserIndex @目前只能显示除自己手牌外的三家，所以UserIndex.down不支持
---@param list_cards table @{11,12,13,14,...}牌值的列表，如果是牌墙，值全部填InvalidValue，如果传nil，则不改变手牌，只刷新显示
---@param int_show_count int @可省略参数，表示在手牌长度为int_count的基础上，显示前几张手牌，用于开场起牌动画
function cards_3d_layout:setHandCards(e_index, list_cards, int_show_count)
    self:insertLog("setHandCards", e_index, list_cards, int_show_count)
    if list_cards then
        local list_cpy = clone(list_cards)
        -- 逻辑上设置手牌
        cards_3d_logic:setCards(e_index, LayoutCardsType.hand, list_cpy)
    end
    self:refreshLayout(e_index, LayoutCardsType.hand, {show_count=int_show_count})
end

-- 一次性设置出牌堆
---@param e_index UserIndex
---@param list_cards table @{11,12,13...}
function cards_3d_layout:setOutCards(e_index, list_cards)
    self:insertLog("setOutCards", e_index, list_cards)
    -- 逻辑上设置出牌堆
    local list_cpy = clone(list_cards)
    cards_3d_logic:setCards(e_index, LayoutCardsType.out, list_cpy)
    self:refreshLayout(e_index, LayoutCardsType.out)
end

-- 一次性设置吃碰杠牌堆
---@param e_index UserIndex
---@param list_heaps table @{{_type=HandleType.b_peng, _value=11, _from=UserIndex.up},{_type=HandleType.b_peng, _value=12, _from=UserIndex.left}...}
function cards_3d_layout:setHeapCards(e_index, list_heaps)
    self:insertLog("setHeapCards", e_index, list_heaps)
    -- 逻辑上设置吃碰杠牌堆
    local h = {}
    for i=1,#list_heaps do
        h[i]={}
        h[i].value=list_heaps[i]._value
        h[i].type=list_heaps[i]._type
        h[i].from=list_heaps[i]._from
    end
    cards_3d_logic:setHeaps(e_index, h)
    self:refreshLayout(e_index, LayoutCardsType.heap)
end

-- 一次性设置补花牌堆
---@param e_index UserIndex
---@param list_cards table
function cards_3d_layout:setHuaCards(e_index, list_cards)
    self:insertLog("setHuaCards", e_index, list_cards)
    -- 逻辑上设置花牌堆
    local list_cpy = clone(list_cards)
    cards_3d_logic:setCards(e_index, LayoutCardsType.hua, list_cpy)
    self:refreshLayout(e_index, LayoutCardsType.hua)
end

function cards_3d_layout:getLastHandCardInfo(e_index)
    self:insertLog("getLastHandCardInfo", e_index)
    local n = cards_3d_logic:getCardsLength(e_index, LayoutCardsType.hand)
    if n > 0 then
        local c = self._out_cards[e_index][n]
        if c == nil then
            self:addOutCard(e_index, InvalidValue, n)
            if self._out_cards[e_index][n] then
                self._out_cards[e_index][n]:hideCard()
            end
            c = self._out_cards[e_index][n]
        end

        if not c then
            -- 超过手牌配置最大范围
            return {x=1234567,y=1234567,index=0}
        end

        local tbl = {}
        tbl.x = c:getPositionX() + c._card_img:getPositionX()
        tbl.y = c:getPositionY() + c._card_img:getPositionY()
        tbl.index = n
        return tbl
    end

    return {x=1234567,y=1234567,index=0}
end

---@param buttonIndex HandleButtonIndex @操作序号
function cards_3d_layout:getHandleSelectButton(buttonIndex)
    if self._handle_select and self._handle_select._btn_handle then
        return self._handle_select._btn_handle[buttonIndex]
    end
    return nil
end

-- 设置给手牌排序的callback
---@param cb_function function @Sample:
---cards_3d_layout:setSortCardsFunc(function({11,11,13,12,11,12,13,...})
---    return {11,11,11,12,12,13,13,...}
---end)
function cards_3d_layout:setSortCardsFunc(cb_function)
    self:insertLog("setSortCardsFunc", nil)
    if cb_function and type(cb_function) == "function" then
        self._sort_cards_func = cb_function
    end
end
-- 设置出牌动画结束的callback，出牌动画结束以后会调用，一般用来显示光标
---@param cb_function function @Sample:
---cards_3d_layout:setOutCardEndFunc(function(e_index, pos)
---    self:showCardPoint(e_index, pos)
---    ...
---end)
function cards_3d_layout:setOutCardEndFunc(cb_function)
    self:insertLog("setOutCardEndFunc", nil)
    if cb_function and type(cb_function) == "function" then
        self._outcard_end_func = cb_function
    end
end

-- 设置如何显示某张暗杠的callback，如果不设置，则根据LayoutConfig显示
---@param cb_function function @Sample:
---cards_3d_layout:setOnConfigDarkGangFunc(function(card_value)
---    if card_value == 11 then
---        return { DarkGangConfig.Back,
---                 DarkGangConfig.Back,
---                 DarkGangConfig.Back,
---                 DarkGangConfig.Value
---        }
---    end
---    return {}
---end)
function cards_3d_layout:setOnConfigDarkGangFunc(cb_function)
    self:insertLog("setOnConfigDarkGangFunc", nil)
    if cb_function and type(cb_function) == "function" then
        self._on_config_dark_gang_func = cb_function
    end
end

-- 设置框架有任意输入时会调用的callback，一般用来关闭系统菜单等
---@param cb_function function @Sample:
---cards_3d_layout:setOnAnyTouchEventFunc(function(sender,event)
---    ...
---end)
function cards_3d_layout:setOnAnyTouchEventFunc(cb_function)
    self:insertLog("setOnAnyTouchEventFunc", nil)
    if cb_function and type(cb_function) == "function" then
        self._on_any_touch_event_func = cb_function
    end
end

-- 设置把自己手牌拖动到屏幕上方放开之后会调用的callback
---@param cb_function function @Sample:
---cards_3d_layout:setMoveCardOutFunc(function(byte_card)
---    local card_value = byte_card
---    ...
---end)
function cards_3d_layout:setMoveCardOutFunc(cb_function)
    self:insertLog("setMoveCardOutFunc", nil)
    if cb_function and type(cb_function) == "function" then
        self._move_card_out_func = cb_function
    end
end

-- 设置把点击手牌放大，再点击一次之后会调用的callback
---@param cb_function function @Sample:
---cards_3d_layout:setClickCardOutFunc(function(byte_card)
---    local card_value = byte_card
---    ...
---end)
function cards_3d_layout:setClickCardOutFunc(cb_function)
    self:insertLog("setClickCardOutFunc", nil)
    if cb_function and type(cb_function) == "function" then
        self._click_card_out_func = cb_function
    end
end

-- 设置点击手牌放大之后会调用的callback
---@param cb_function function @Sample:
---cards_3d_layout:setClickCardFunc(function(byte_card)
---    cards_3d_layout:setDarkCardValue(byte_card)
---end)
function cards_3d_layout:setClickCardFunc(cb_function)
    self:insertLog("setClickCardFunc", nil)
    if cb_function and type(cb_function) == "function" then
        self._click_card_func = cb_function
    end
end

-- 设置插牌位置处理函数，传入参数是除插入牌之外的牌值列表，和要插入的牌值，返回要插入的位置序号
---@param cb_function function @Sample:
---cards_3d_layout:setFindCardInHandPositionFunc(function(cards,value)
---    local current_hand_cards = cards -- {11,12,13,14...}
---    local card_value = value -- 11..19|21..29|...
---    ...
---end)
function cards_3d_layout:setFindCardInHandPositionFunc(cb_function)
    if cb_function and type(cb_function) == "function" then
        self._find_card_in_hand_position_func = cb_function
        CSelfCardsController:setFindCardInHandPositionFunc(function(cards,value)
            return self._find_card_in_hand_position_func(cards,value)
        end)
    end
end

-- 设置Handle选择面板点击后的回调callback，被选择的操作以参数形式传递
---@param cb_function function @Sample:
---cards_3d_layout:setHandleClickedFunc(function(arg)
---    local t = arg.handle_type --HandleType.pass|HandleType.a_gang|...
---    local v = arg.card_value --11..19|21..29|...
---end)
function cards_3d_layout:setHandleClickedFunc(cb_function)
    self:insertLog("setHandleClickedFunc", nil)
    if cb_function and type(cb_function) == "function" then
        self._handle_clicked_func = cb_function
    end
end

---@param bool_can_touch bool @是否响应点击
---@param optional list_index table @{1,3,4,5,6...}哪些手牌会被设置，参数是序号组成的table，可以不传，不传则全部设置
function cards_3d_layout:setHandCardsCanTouch(bool_can_touch, list_index)
    self:insertLog("setHandCardsCanTouch", bool_can_touch, list_index)
    local list_cpy = clone(list_index)
    CSelfCardsController:setHandCardsCanTouch(bool_can_touch, list_cpy)
end

---@param bool_can_touch bool @是否响应点击
---@param optional list_value table @{1,3,4,5,6...}牌值列表
function cards_3d_layout:setHandCardsCanTouchByValue(bool_can_touch, list_value)
    self:insertLog("setHandCardsCanTouchByValue", bool_can_touch, list_value)
    local list_cpy = clone(list_value)
    CSelfCardsController:setHandCardsCanTouchByValue(bool_can_touch, list_cpy)
end

---@param bool_can_out bool @是否可以被打出
---@param optional list_index table @{1,3,4,5,6...}哪些手牌会被设置，参数是序号组成的table，可以不传，不传则全部设置
function cards_3d_layout:setHandCardsCanOut(bool_can_out, list_index)
    self:insertLog("setHandCardsCanOut", bool_can_out, list_index)
    local list_cpy = clone(list_index)
    CSelfCardsController:setHandCardsCanOut(bool_can_out, list_cpy)
end

---@param bool_can_out bool @是否可以被打出
---@param optional list_value table @{1,3,4,5,6...}牌值列表
function cards_3d_layout:setHandCardsCanOutByValue(bool_can_out, list_value)
    self:insertLog("setHandCardsCanOutByValue", bool_can_out, list_value)
    local list_cpy = clone(list_value)
    CSelfCardsController:setHandCardsCanOutByValue(bool_can_out, list_cpy)
end

-- 通过指定手牌序号来指定哪些手牌要变黑
---@param bool_set_dark bool @是否变黑
---@param optional list_index table @{1,3,4,5,6...}哪些手牌会被设置，参数是序号组成的table，可以不传，不传则全部设置
function cards_3d_layout:setHandCardsDark(bool_set_dark, list_index)
    self:insertLog("setHandCardsDark", bool_set_dark, list_index)
    local list_cpy = clone(list_index)
    CSelfCardsController:setHandCardsDark(bool_set_dark, list_cpy)
end

-- 通过指定手牌牌值来指定哪些手牌要变黑
---@param bool_set_dark bool @是否变黑
---@param optional list_value table @{1,3,4,5,6...}牌值列表
function cards_3d_layout:setHandCardsDarkByValue(bool_set_dark, list_value)
    self:insertLog("setHandCardsDarkByValue", bool_set_dark, list_value)
    local list_cpy = clone(list_value)
    CSelfCardsController:setHandCardsDarkByValue(bool_set_dark, list_cpy)
end

-- 通过手牌序号指定哪些牌显示听牌标记，未被指定的会隐藏，如果要全部隐藏可以传{}或nil
---@param optional list_index table @{1,3,4,5,6...}哪些手牌会被打上听牌标记，其余手牌则不显示标记，参数是序号组成的table，可以不传，不传则全部设置
function cards_3d_layout:setHandCardsTing(list_index)
    self:insertLog("setHandCardsTing", list_index)
    local list_cpy = clone(list_index)
    CSelfCardsController:setHandCardsTing(list_cpy)
end

-- 通过指定手牌值列表来指定哪些手牌需要显示听牌标记，未被指定的会隐藏，如果要全部隐藏可以传{}或nil
---@param optional list_value table @{11,13,14,...}手牌牌值如果在list里会被打上听牌标记
function cards_3d_layout:setHandCardsTingByValue(list_value)
    self:insertLog("setHandCardsTingByValue", list_value)
    local list_cpy = clone(list_value)
    CSelfCardsController:setHandCardsTingByValue(list_cpy)
end

function cards_3d_layout:setHandCardsTingDaByValue(list_value)
    self:insertLog("setHandCardsTingDaByValue", list_value)
    local list_cpy = clone(list_value)
    CSelfCardsController:setHandCardsTingDaByValue(list_cpy)
end

function cards_3d_layout:setHandCardsTingDuoByValue(list_value)
    self:insertLog("setHandCardsTingDuoByValue", list_value)
    local list_cpy = clone(list_value)
    CSelfCardsController:setHandCardsTingDuoByValue(list_cpy)
end

--
function cards_3d_layout:setLayIndex(index)
    CSelfCardsController:setLayIndex(index);
end

--
function cards_3d_layout:setSelectType(type)
    CSelfCardsController:setSelectedIndex(0)
    CSelfCardsController:setSelectedTable({});
    CSelfCardsController:refreshHandCards()
    CSelfCardsController:setSelectType(type);
end

--
function cards_3d_layout:setSelfHandJump(index,interval,high,times,lock,unlock)
    CSelfCardsController:setSelfHandJump(index,interval,high,times,lock,unlock)
end

-- 获取多选牌
function cards_3d_layout:getSelectTable()
    local table = CSelfCardsController:getSelectedTable();    
    return table;
end 
-- 获取多选牌Index
function cards_3d_layout:getSelectIndexTable()
    local table = CSelfCardsController:getSelectIndexTable()
    return table;
end

-- 选起多张手牌
function cards_3d_layout:selectMoveHandCards(table_index)
    CSelfCardsController:setSelectedIndex(0)
    CSelfCardsController:setSelectedTable(table_index);
    CSelfCardsController:refreshHandCards()
end

-- 选起多张手牌
function cards_3d_layout:selectMoveHandCardsWithValue(table_value)
    CSelfCardsController:setSelectedIndex(0)
    CSelfCardsController:setSelectedValue(table_value);
    CSelfCardsController:refreshHandCards()
end

-- 取消手牌的选择状态
function cards_3d_layout:unselectAllHandCards()
    CSelfCardsController:setSelectedIndex(0)
    CSelfCardsController:setSelectedTable({});
    CSelfCardsController:refreshHandCards()
end

-- 某玩家抓了一张牌，插入手牌
---@param e_index UserIndex
---@param byte_card int
function cards_3d_layout:catchCard(e_index, byte_card)
    self:insertLog("catchCard", e_index, byte_card)
    -- 逻辑上插入手牌
    cards_3d_logic:catchCard(e_index, byte_card)
    self:refreshLayout(e_index, LayoutCardsType.hand)
end

-- 某玩家出了一张手牌到出牌堆，如果不设起始位置，默认位置是对应方位的玩家的最后一张手牌
---@param e_index UserIndex
---@param byte_card int
---@param optional position @{x=100,y=200} 自己出牌才需要的参数，出牌的位置坐标
---@param optional int_pos @自己出牌才需要的参数，出的牌是自己手牌的第几张
function cards_3d_layout:outCard(e_index, byte_card, position, int_pos)
    self:insertLog("outCard", e_index, byte_card, position, int_pos)
    -- 在逻辑变化前先获取最后一张手牌作为起始点
    local begin_info = self:getLastVisibleCardInfo(e_index, LayoutCardsType.hand)
    if position then
        begin_info = {}
        begin_info.x = position.x
        begin_info.y = position.y
    end

    if LayoutConfig.play_insert_card and e_index == UserIndex.down then
        -- 自己的手牌出牌，在逻辑上被改变之前，播放插牌动画
        CSelfCardsController._hand_cards[int_pos]:hideCard()
        CSelfCardsController:insertLastCardInHandCards(int_pos)
    end
    -- 逻辑上出牌
    cards_3d_logic:outCard(e_index, byte_card)

    if LayoutConfig.play_insert_card and e_index == UserIndex.down then
    else
        -- 如果是自己的手牌，不直接刷新，动画放完之后会刷新
        self:refreshLayout(e_index, LayoutCardsType.hand)
    end
    self:refreshLayout(e_index, LayoutCardsType.out)

    -- 获取出牌的目标位置
    local visual_end_info = self:getLastVisibleCardInfo(e_index, LayoutCardsType.out)
    -- 出牌堆牌的真实位置和显示位置有一个offset，在getLastVisibleCardInfo里已经加上了，在这里要减掉
    local c = self._out_cards[e_index][visual_end_info.index]
    local end_position = ccp(visual_end_info.x - c._card_img:getPositionX(), visual_end_info.y - c._card_img:getPositionY())

    -- 以下是显示从出牌位置到出牌堆移动的动画

    ---- 没有找到合适的起始信息，并且没有传入起始位置参数，不显示动画
    if not begin_info then
        if self._outcard_end_func then
            self._outcard_end_func(e_index, ccp(visual_end_info.x, visual_end_info.y))
        end
        return
    end

    local move_array=CCArray:create()

    -- 贝塞尔曲线设置
    local bezier=ccBezierConfig:new()
    -- 控制点1(起点)
    if e_index == UserIndex.up or e_index == UserIndex.down then
        bezier.controlPoint_1 = ccp(begin_info.x, end_position.y)
        bezier.controlPoint_2 = ccp(begin_info.x, end_position.y)
    elseif e_index == UserIndex.left or e_index == UserIndex.right then
        bezier.controlPoint_1 = ccp(end_position.x, begin_info.y)
        bezier.controlPoint_2 = ccp(end_position.x, begin_info.y)
    end

    -- 终点，要移动到的位置是真实位置，所以用end_position
    bezier.endPosition = ccp(end_position.x, end_position.y)
    print("bezier end:x:"..end_position.x.." y:"..end_position.y);
    local tmp_zorder = self._out_cards[e_index][visual_end_info.index]:getZOrder()
    move_array:addObject(CCBezierTo:create(0.2, bezier))
    move_array:addObject(CCCallFunc:create(function()
        self._out_cards[e_index][visual_end_info.index]:setZOrder(tmp_zorder)
        if self._outcard_end_func then
            self._outcard_end_func(e_index, ccp(visual_end_info.x, visual_end_info.y))
        end
        frame_tools:restore_message()
    end))
    local move_action=CCSequence:create(move_array)

    frame_tools:pause_message()
    self._out_cards[e_index][visual_end_info.index]:setPosition(ccp(begin_info.x, begin_info.y))
    self._out_cards[e_index][visual_end_info.index]:setZOrder(LayoutZOrder.Topest)
    self._out_cards[e_index][visual_end_info.index]:runAction(move_action)
end

-- 某玩家出了一张手牌到补花牌堆
---@param e_index UserIndex
---@param byte_card int
function cards_3d_layout:buHua(e_index, byte_card)
    self:insertLog("buHua", e_index, byte_card)
    -- 逻辑上补花
    cards_3d_logic:buHua(e_index, byte_card)
    -- 补花结束强制给手牌排序
    cards_3d_logic:sortHandCards(e_index)
    self:refreshLayout(e_index, LayoutCardsType.hand)
    self:refreshLayout(e_index, LayoutCardsType.hua)
end

-- 某玩家吃碰杠了某玩家打出的牌，除了暗杠，必须指定from
---@param e_index UserIndex
---@param e_kind HandleType
---@param e_from_index UserIndex
function cards_3d_layout:handle(e_index, e_from_index, e_kind, byte_value)
    self:insertLog("handle", e_index, e_from_index, e_kind, byte_value)

    if e_index == UserIndex.down and (e_kind == HandleType.d_gang or e_kind == HandleType.a_gang) then
        ---- 暗杠，补杠的时候，不播放牌插入动画，要先整理手牌
        --local cards = cards_3d_logic:getCards(UserIndex.down, LayoutCardsType.hand)
        --local card_value = cards[#cards]
        --local index = #cards
        --if self._find_card_in_hand_position_func then
        --    index = self._find_card_in_hand_position_func(cards, card_value)
        --else
        --    index = CSelfCardsController:defaultFindCardPosition(cards, card_value, index)
        --end
        --cards_3d_logic:moveCard(UserIndex.down, LayoutCardsType.hand, #cards, index)

        -- 直接用排序
        cards_3d_logic:sortHandCards(e_index)
    end

    -- 逻辑上吃碰杠
    cards_3d_logic:handle(e_index, e_from_index, e_kind, byte_value)
    self:refreshLayout(e_index, LayoutCardsType.hand)
    self:refreshLayout(e_index, LayoutCardsType.heap)
    if e_from_index and e_from_index ~= -1 then
        self:refreshLayout(e_from_index, LayoutCardsType.out)
    end
end

-- 设置桌上使用灰色遮罩的牌值，如果不需要，传0或nil
---@param byte_card int
function cards_3d_layout:setDarkCardValue(byte_card)
    self:insertLog("setDarkCardValue", byte_card)
    self._dark_card_value = byte_card
    CSelfCardsController:setDarkCardValue(byte_card)
    self:refreshLayout(UserIndex.down, LayoutCardsType.out)
    self:refreshLayout(UserIndex.up, LayoutCardsType.out)
    self:refreshLayout(UserIndex.left, LayoutCardsType.out)
    self:refreshLayout(UserIndex.right, LayoutCardsType.out)
    self:refreshLayout(UserIndex.down, LayoutCardsType.heap)
    self:refreshLayout(UserIndex.up, LayoutCardsType.heap)
    self:refreshLayout(UserIndex.left, LayoutCardsType.heap)
    self:refreshLayout(UserIndex.right, LayoutCardsType.heap)
    self:refreshLayout(UserIndex.down, LayoutCardsType.hua)
    self:refreshLayout(UserIndex.up, LayoutCardsType.hua)
    self:refreshLayout(UserIndex.left, LayoutCardsType.hua)
    self:refreshLayout(UserIndex.right, LayoutCardsType.hua)
    --
    --self:refreshSelfCards(LayoutCardsType.heap)
end

-- 设置有角标的牌值
---@param config table @config[47]=1;config[41]=2;...
function cards_3d_layout:setCornerCardValue(config)
    self:insertLog("setCornerCardValue", config)
    local list_cpy = clone(config)
    self._corner_config = list_cpy
    CSelfCardsController:setCornerCardValue(list_cpy)
    -- 角标设完不刷新界面，否则会在游戏开始显示所有手牌，如果需要，手动刷新
    --self:refreshLayout(UserIndex.down, LayoutCardsType.out)
    --self:refreshLayout(UserIndex.up, LayoutCardsType.out)
    --self:refreshLayout(UserIndex.left, LayoutCardsType.out)
    --self:refreshLayout(UserIndex.right, LayoutCardsType.out)
    --self:refreshLayout(UserIndex.down, LayoutCardsType.heap)
    --self:refreshLayout(UserIndex.up, LayoutCardsType.heap)
    --self:refreshLayout(UserIndex.left, LayoutCardsType.heap)
    --self:refreshLayout(UserIndex.right, LayoutCardsType.heap)
    --self:refreshLayout(UserIndex.down, LayoutCardsType.hand)
    --self:refreshLayout(UserIndex.up, LayoutCardsType.hand)
    --self:refreshLayout(UserIndex.left, LayoutCardsType.hand)
    --self:refreshLayout(UserIndex.right, LayoutCardsType.hand)
end

-- 显示或隐藏蓝色小箭头
---@param e_index UserIndex @要显示蓝色小箭头的方位，传nil或0则不显示
function cards_3d_layout:showLastOutCardIndicator(e_index)
    if not self._last_out_card_indicator then
        self._last_out_card_indicator = CCardPoint3D.new()
        self._view_root:addNode(self._last_out_card_indicator)
        self._last_out_card_indicator:setZOrder(LayoutZOrder.LastOutCardIndicator)
    end
    if e_index and e_index ~= 0 then
        local info = self:getLastVisibleCardInfo(e_index, LayoutCardsType.out)
        if info then
            self._last_out_card_indicator:setPosition(ccp(info.x, info.y))
            self._last_out_card_indicator:showPoint()
        end
    else
        self._last_out_card_indicator:hidePoint()
    end
end

-- 设置handle数据，顺序按照传入参数的顺序，如果需要“过”显示在最左边，则第一个参数就传HandleType.pass
---@param list_handle table @{{type=HandleType.b_chi_left, value=11},{type=HandleType.b_gang, value=35}...}
function cards_3d_layout:showHandleSelect(list_handle)
    self:insertLog("showHandleSelect", list_handle)
    local list_cpy = clone(list_handle)
    self._handle_select:setHandleData(list_cpy)
    self._handle_select:show()
end

function cards_3d_layout:hideHandleSelect()
    self:insertLog("hideHandleSelect", nil)
    self._handle_select:hide()
end

--显示牌遮罩
function cards_3d_layout:showCardsMask(e_index,isShow)
    if e_index == UserIndex.down then return; end

    self._card_mask[e_index] = isShow;
    self:refreshLayout(e_index, LayoutCardsType.hand)
    self:refreshLayout(e_index, LayoutCardsType.heap)
end

---@param e_index UserIndex
---@param e_type HandleType
---@param byte_card int
function cards_3d_layout:playHandleAnimation(e_index,e_type,byte_card)
    self:insertLog("playHandleAnimation", e_index,e_type,byte_card)
    if not self._handle_ani[e_index] then
        self._handle_ani[e_index]=CCArmature:create("dh_cpght")
        self._handle_ani[e_index]:setPosition(self._handle_ani_pos[e_index])
        self._handle_ani[e_index]:setZOrder(LayoutZOrder.HandleAni)
        self._view_root:addNode(self._handle_ani[e_index])

        self._handle_ani[e_index].bone1 = self._handle_ani[e_index]:getBone("Layer12")
        self._handle_ani[e_index].bone2 = self._handle_ani[e_index]:getBone("Layer13")
        self._handle_ani[e_index].bone3 = self._handle_ani[e_index]:getBone("Layer14")
        self._handle_ani[e_index].bone4 = self._handle_ani[e_index]:getBone("Layer4")
    end
    local array=CCArray:create()
    self._handle_ani[e_index]:setVisible(true)
    local show_time = 0.8
    --local bone_card_scale = 66 / HandCardWidth -- 动画资源里宽度是66
    --TODO: check scale
    self._card_imgs = {}
    local bone_card_scale = 1
    if e_type==HandleType.b_peng then
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone1, byte_card, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone2, byte_card, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone3, byte_card, bone_card_scale))
        self._handle_ani[e_index]:getAnimation():play("dh_peng",-1,-1,-1,-1)
        array:addObject(CCDelayTime:create(show_time))
        array:addObject(CCCallFunc:create(function() self:commonHandleAniFinished(e_index) end))
    elseif e_type==HandleType.b_chi_left then
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone1, byte_card, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone2, byte_card+1, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone3, byte_card+2, bone_card_scale))
        self._handle_ani[e_index]:getAnimation():play("dh_chi",-1,-1,-1,-1)
        array:addObject(CCDelayTime:create(show_time))
        array:addObject(CCCallFunc:create(function() self:commonHandleAniFinished(e_index) end))
    elseif e_type==HandleType.b_chi_mid then
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone1, byte_card-1, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone2, byte_card, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone3, byte_card+1, bone_card_scale))
        self._handle_ani[e_index]:getAnimation():play("dh_chi",-1,-1,-1,-1)
        array:addObject(CCDelayTime:create(show_time))
        array:addObject(CCCallFunc:create(function() self:commonHandleAniFinished(e_index) end))
    elseif e_type == HandleType.b_chi_right then
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone1, byte_card-2, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone2, byte_card-1, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone3, byte_card, bone_card_scale))
        self._handle_ani[e_index]:getAnimation():play("dh_chi",-1,-1,-1,-1)
        array:addObject(CCDelayTime:create(show_time))
        array:addObject(CCCallFunc:create(function() self:commonHandleAniFinished(e_index) end))
    elseif e_type==HandleType.b_gang then
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone1, byte_card, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone2, byte_card, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone3, byte_card, bone_card_scale))
        table.insert(self._card_imgs, self:replayBoneCard(self._handle_ani[e_index].bone4, byte_card, bone_card_scale))
        self._handle_ani[e_index]:getAnimation():play("dh_gang",-1,-1,-1,-1)
        array:addObject(CCDelayTime:create(show_time))
        array:addObject(CCCallFunc:create(function() self:commonHandleAniFinished(e_index) end))
    elseif e_type==HandleType.d_gang or e_type==HandleType.a_gang then
        self._handle_ani[e_index]:getAnimation():play("dh_gang_an",-1,-1,-1,-1)
        array:addObject(CCDelayTime:create(show_time))
        array:addObject(CCCallFunc:create(function() self:commonHandleAniFinished(e_index) end))
    elseif e_type==HandleType.hu then
        self._handle_ani[e_index]:getAnimation():play("dh_hu",-1,-1,-1,-1)
        array:addObject(CCDelayTime:create(2))
    elseif e_type==HandleType.ting then
        self._handle_ani[e_index]:getAnimation():play("dh_ting",-1,-1,-1,-1)
        array:addObject(CCDelayTime:create(2))
    end

    local action=CCSequence:create(array)
    self._handle_ani[e_index]:stopAllActions()
    self._handle_ani[e_index]:runAction(action)

    -- 牌面是牌底的子节点，现在集成的tolua并没有导出setCascadeOpacityEnabled这个函数
    -- 只能自己设定一个时间，让牌面fadeout
    if self._time_schedule then
        CCDirector:sharedDirector():getScheduler():unscheduleScriptEntry(self._time_schedule)
        self._time_schedule=nil
    end
    -- 设定fadeout时间
    self._card_fade_out_time = 0.8
    -- 建立一个tick来处理透明度
    self._time_schedule=CCDirector:sharedDirector():getScheduler():scheduleScriptFunc(function(dt) self:scheduleTimer(dt) end , 0.1,false)
end

------------------------------------------------------------------------------------------------------------------
-- 以下接口原则上是private的
------------------------------------------------------------------------------------------------------------------

-- 获取table的字符串格式内容，递归
---@param tab table
---@param ind number @不用传此参数，递归用（前缀格式（空格））
---@return string @format string of the table
function cards_3d_layout:dumpTab(tab,ind)
    if(tab==nil)then return "nil" end;
    local str="{";
    if(ind==nil)then ind="  "; end;
    --//each of table
    for k,v in pairs(tab) do
        --//key
        if(type(k)=="string")then
            k=tostring(k).." = ";
        else
            k="["..tostring(k).."] = ";
        end;--//end if
        --//value
        local s="";
        if(type(v)=="nil")then
            s="nil";
        elseif(type(v)=="boolean")then
            if(v) then s="true"; else s="false"; end;
        elseif(type(v)=="number")then
            s=v;
        elseif(type(v)=="string")then
            s="\""..v.."\"";
        elseif(type(v)=="table")then
            s=self:dumpTab(v,ind.."  ");
            s=string.sub(s,1,#s-1);
        elseif(type(v)=="function")then
            s="function : "..tostring(v);
        elseif(type(v)=="thread")then
            s="thread : "..tostring(v);
        elseif(type(v)=="userdata")then
            s="userdata : "..tostring(v);
        else
            s="nuknow : "..tostring(v);
        end;--//end if
        --//Contact
        str=str.."\n"..ind..k..s.." ,";
    end --//end for
    --//return the format string
    local sss=string.sub(str,1,#str-1);
    if(#ind>0)then ind=string.sub(ind,1,#ind-2) end;
    sss=sss.."\n"..ind.."}\n";
    return sss;--string.sub(str,1,#str-1).."\n"..ind.."}\n";
end;--//end function

-- 插入一条log
---@param fun_name string @调用函数名
---@param ... table @所有参数
function cards_3d_layout:insertLog(fun_name, ...)
    if LayoutConfig.log_enable or LayoutConfig.print_enable then
        if not self._log then
            self._log = ""
        end
        local node = "cards_3d_layout:"..tostring(fun_name)..":\t"
        if {...} then
            node = node..self:dumpTab({...})
        end
        node = node.."\n"
        if LayoutConfig.print_enable then
            print(node)
        end
        if LayoutConfig.log_enable then
            self._log = self._log..node
        end
    end
end

-- 这个函数只用来处理吃碰杠动画中的牌面渐隐
---@param dt number @deltaT
function cards_3d_layout:scheduleTimer(dt)
    self._card_fade_out_time=self._card_fade_out_time-dt
    if self._card_fade_out_time<0 then
        CCDirector:sharedDirector():getScheduler():unscheduleScriptEntry(self._time_schedule)
        self._time_schedule=nil

        -- 时间到，设不可见
        for i=1,#self._card_imgs do
            local card = self._card_imgs[i]
            card:setVisible(false)
        end
        return
    end

    -- 根据剩余时间来设定自己的不透明度
    for i=1,#self._card_imgs do
        local card = self._card_imgs[i]
        card:setOpacity(self._card_fade_out_time/0.8 * 255)
    end
end

-- 生成一张对应牌值的牌，替换掉骨骼节点
---@param bone Bone @动画骨骼节点
---@param byte_card int @牌值
---@param card_scale float @缩放
---@return card @生成的牌面，该牌面会被保存到一个table里，在tick渐隐
function cards_3d_layout:replayBoneCard(bone, byte_card, card_scale)
    -- 替换财神和白板
    local card_value = byte_card
    if CardExchangeTable[card_value] ~= nil then
        card_value = CardExchangeTable[card_value]
    end
    local card = CSelfCardsController:newHandCard()
    local card_info,base_x,base_y,base_zorder = CSelfCardsController:makeHandCardLayoutInfo(1, true)
    card:init(card_info,base_x,base_y,base_zorder)
    card:setPosition(ccp(0,0))
    card:setScale(card_scale)
    card:showCard(card_value)
    bone:addDisplay(card,0)

    return card._card_img
end

-- handle动画播放完毕
---@param e_index UserIndex
function cards_3d_layout:commonHandleAniFinished(e_index)
    self._handle_ani[e_index]:setVisible(false)
end

-- 获取逻辑控制器
---@return cards_3d_logic @牌控制逻辑
function cards_3d_layout:getLogic()
    return cards_3d_logic
end

-- 刷新自己的手牌和吃碰杠牌堆
---@param e_type LayoutCardsType
---@param args table @{show_count=4}
function cards_3d_layout:refreshSelfCards(e_type, args)
    if e_type == LayoutCardsType.hand then
        local cards = cards_3d_logic:getCards(UserIndex.down, e_type)
        local int_show_count = #cards
        if args and args.show_count then
            int_show_count = args.show_count
        end
        CSelfCardsController:setHandCards(cards, int_show_count)
    elseif e_type == LayoutCardsType.heap then
        local heaps = cards_3d_logic:getCards(UserIndex.down, e_type)
        CSelfCardsController:setHeapCards(heaps)
    end
end

---@param bool_lay bool @是否摊开自己手牌
function cards_3d_layout:laySelfHandCards(bool_lay)
    CSelfCardsController:layHandCards(bool_lay)
end

---@param bool_lay bool @是否摊开
---@return card @生成一张手牌
function cards_3d_layout:getNewHandCard(bool_lay)
    local card = CSelfCardsController:newHandCard()
    local card_info,base_x,base_y,base_zorder = CSelfCardsController:makeHandCardLayoutInfo(1, bool_lay)
    card:init(card_info,base_x,base_y,base_zorder)
    card:setPosition(ccp(0,0))
    return card
end

-- 刷新界面，某方位的某牌堆的所有牌（除了自己手牌和自己吃碰杠牌堆）
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@param args table @{show_count=4}
function cards_3d_layout:refreshLayout(e_index, e_type, args)

    if e_index == UserIndex.down then
        -- 自己的手牌或吃碰杠牌，用独立逻辑处理
        if e_type == LayoutCardsType.hand or e_type == LayoutCardsType.heap then
            self:refreshSelfCards(e_type, args)
            return
        end
    end

    if e_type == LayoutCardsType.hand and not self:isSelfIndex(e_index) then
        -- 手牌，忽略自己的手牌
        local cards = cards_3d_logic:getCards(e_index, e_type)
        local int_count = #cards
        local int_show_count = int_count
        if args and args.show_count and args.show_count ~= 0 then
            int_show_count = args.show_count
        end

        local item_name
        if int_count < 3 then
            item_name = "items_2"
        elseif int_count < 6 then
            item_name = "items_5"
        elseif int_count < 9 then
            item_name = "items_8"
        elseif int_count < 12 then
            item_name = "items_11"
        elseif int_count < 15 then
            item_name = "items_14"
        elseif int_count < 18 then
            item_name = "items_17"
        end
        
        if item_name == nil then
            if write_online_logs then
                write_online_logs("count:"..tostring(int_count));
            end
            item_name = "items_17"
        end

        for i = 1, int_show_count do
            self:addHandCard(e_index, item_name, cards[i], i)
        end
        for i = #self._hand_cards[e_index], int_show_count+1, -1 do
            self:removeHandCard(e_index, i)
        end
    elseif e_type == LayoutCardsType.out then
        -- 出牌堆
        local cards = cards_3d_logic:getCards(e_index, e_type)
        for i = 1, #cards do
            self:addOutCard(e_index, cards[i], i)
        end
        for i = #self._out_cards[e_index], #cards+1, -1 do
            self:removeOutCard(e_index, i)
        end
    elseif e_type == LayoutCardsType.heap and not self:isSelfIndex(e_index) then
        -- 吃碰杠牌堆，忽略自己的
        local heaps = cards_3d_logic:getCards(e_index, e_type)
        for i = 1, #heaps do
            self:addHeapCard(e_index, heaps[i].from, heaps[i].type, heaps[i].value, i)
        end
        for i = #self._heap_cards[e_index], #heaps+1, -1 do
            self:removeHeapCard(e_index, i)
        end
    elseif e_type == LayoutCardsType.hua then
        -- 花牌堆
        local cards = cards_3d_logic:getCards(e_index, e_type)
        if LayoutConfig.tile_hua_card then
            -- 花牌堆叠模式
            local count_list = {}
            local final_cards = {}
            -- 统计相同花牌数量
            for i=1,#cards do
                if count_list[cards[i]] then
                    count_list[cards[i]] = count_list[cards[i]] + 1
                else
                    count_list[cards[i]] = 1
                    table.insert(final_cards, cards[i])
                end
            end
            -- 一共有count张不同的花牌
            for i=1,#final_cards do
                self:addHuaCard(e_index, final_cards[i], i)
                local count = count_list[final_cards[i]]
                if count > 4 then count = 4 end
                if count < 1 then count = 1 end
                if count > 1 then
                    self._hua_cards[e_index][i]:showCorner(count+2)
                else
                    self._hua_cards[e_index][i]:hideCorner()
                end
            end
            for i = #self._hua_cards[e_index], #final_cards+1, -1 do
                self:removeHuaCard(e_index, i)
            end
        else
            -- 普通模式
            for i = 1, #cards do
                self:addHuaCard(e_index, cards[i], i)
            end
            for i = #self._hua_cards[e_index], #cards+1, -1 do
                self:removeHuaCard(e_index, i)
            end
        end
    end
end

-- 获取[某一方位玩家][某牌堆]在桌上可见的最后一张牌的信息
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@return info @{x=10,y=10,index=1}
function cards_3d_layout:getLastVisibleCardInfo(e_index, e_type)
    local target
    if e_type == LayoutCardsType.hand then
        target = self._hand_cards[e_index]
    elseif e_type == LayoutCardsType.heap then
        target = self._heap_cards[e_index]
    elseif e_type == LayoutCardsType.out then
        target = self._out_cards[e_index]
    elseif e_type == LayoutCardsType.hua then
        target = self._hua_cards[e_index]
    end

    local n = cards_3d_logic:getCardsLength(e_index, e_type)

    local ret_tbl
    if n > 0 then
        for i=n,1,-1 do
            if target[i] and target[i]:isVisible() then
                ret_tbl = {}
                ret_tbl.x = target[i]:getPositionX()
                ret_tbl.y = target[i]:getPositionY()
                if e_type == LayoutCardsType.out then
                    ret_tbl.x = ret_tbl.x + target[i]._card_img:getPositionX()
                    ret_tbl.y = ret_tbl.y + target[i]._card_img:getPositionY()
                end
                ret_tbl.index = i
                break
            end
        end
    end

    return ret_tbl
end

-- 获取[某一方位玩家][某牌堆]在桌上可见的最后一张牌的信息
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@return info @{x=10,y=10,index=1}
function cards_3d_layout:getLastVisibleCardInfoEX(e_index, e_type, e_value)
    local target
    if e_type == LayoutCardsType.hand then
        target = self._hand_cards[e_index]
    elseif e_type == LayoutCardsType.heap then
        target = self._heap_cards[e_index]
    elseif e_type == LayoutCardsType.out then
        target = self._out_cards[e_index]
    elseif e_type == LayoutCardsType.hua then
        target = self._hua_cards[e_index]
    end

    local n = cards_3d_logic:getCardsLength(e_index, e_type)

    local ret_tbl
    if n > 0 then
        for i=n,1,-1 do
            if target[i] and target[i]:isVisible() and target[i]:getCardValue() == e_value then
                ret_tbl = {}
                ret_tbl.x = target[i]:getPositionX()
                ret_tbl.y = target[i]:getPositionY()
                if e_type == LayoutCardsType.out then
                    ret_tbl.x = ret_tbl.x + target[i]._card_img:getPositionX() - target[i]:getContentSize().width/2;
                    ret_tbl.y = ret_tbl.y + target[i]._card_img:getPositionY() - target[i]:getContentSize().height/2;
                end
                ret_tbl.index = i
                break
            end
        end
    end

    return ret_tbl
end

-- 判断是否是自己的id
---@param e_index UserIndex
---@return success bool
function cards_3d_layout:isSelfIndex(e_index)
    if e_index == UserIndex.down then return true end
    return false
end

-- 添加一张牌到手牌，牌值是byte_card，位置是int_pos可省略，如果位置不是(最大index+1)，原来位置的牌会被替换
---@param e_index UserIndex
---@param s_item string @item name in config according to hand cards number
---@param byte_card int
---@param optional int_pos int
function cards_3d_layout:addHandCard(e_index, s_item, byte_card, int_pos)
    if self:isSelfIndex(e_index) then return end

    local bool_wall = (byte_card == InvalidValue)

    if int_pos == nil then
        int_pos = #self._hand_cards[e_index] + 1
    end

    -- 先new一个新的
    if not self._hand_cards[e_index][int_pos] then
        self._hand_cards[e_index][int_pos] = CHandCard3D.new()
        self._view_root:addChild(self._hand_cards[e_index][int_pos])
    end

    if self._hand_cards[e_index][int_pos] then
        -- 再根据配置初始化
        local word="show"
        if bool_wall then
            word="wall"
        end
        local card_info,base_x,base_y,base_zorder = self:getLayoutInfo(e_index, word, s_item, int_pos)
        self._hand_cards[e_index][int_pos]:init(card_info,base_x,base_y,base_zorder)
        
        self._hand_cards[e_index][int_pos]:showCard(byte_card, self._corner_config[byte_card], false)
        if self._card_mask[e_index] then
            self._hand_cards[e_index][int_pos]:showMask(true);
        end
    end
end

-- 从手牌删除一张牌，位置是int_pos可省略
---@param e_index UserIndex
---@param optional int_pos int
function cards_3d_layout:removeHandCard(e_index, int_pos)
    if self._hand_cards[e_index] and #self._hand_cards[e_index] > 0 then
        if not int_pos then
            local info = self:getLastVisibleCardInfo(e_index, LayoutCardsType.hand)
            int_pos = info.index
        end

        self._hand_cards[e_index][int_pos]:hideCard()
    end
end

-- 添加一张牌到出牌堆，牌值是byte_card，位置是int_pos可省略，如果位置不是(最大index+1)，原来位置的牌会被替换
---@param e_index UserIndex
---@param byte_card int
---@param optional int_pos int
function cards_3d_layout:addOutCard(e_index, byte_card, int_pos)
    if int_pos == nil then
        local info = self:getLastVisibleCardInfo(e_index, LayoutCardsType.out)
        if info then
            int_pos = info.index+1
        else
            int_pos = 1
        end
    end
    if not self._out_cards[e_index][int_pos] then
        local card_info,base_x,base_y,base_zorder = self:getLayoutInfo(e_index, "out", "items", int_pos)
        -- 验证json里的"id"字段是否合法，目前看是顺序排列，从0开始的，如果将来有变化，需要改变逻辑找到正确的id对应的配置
        if card_info.id == int_pos - 1 then
            self._out_cards[e_index][int_pos] = CTiledCard3D.new(card_info,base_x,base_y,base_zorder)
            self._view_root:addChild(self._out_cards[e_index][int_pos])
        end
    end

    if self._out_cards[e_index][int_pos] then
        self._out_cards[e_index][int_pos]:showCard(byte_card, self._corner_config[byte_card], byte_card == self._dark_card_value)
    end
end

-- 从出牌堆删除一张牌，位置是int_pos可省略
---@param e_index UserIndex
---@param optional int_pos int
function cards_3d_layout:removeOutCard(e_index, int_pos)
    if self._out_cards[e_index] and #self._out_cards[e_index] > 0 then
        if not int_pos then
            local info = self:getLastVisibleCardInfo(e_index, LayoutCardsType.out)
            int_pos = info.index
        end
        self._out_cards[e_index][int_pos]:hideCard()
    end
end

-- 添加一组牌到吃碰杠牌堆，类型是牌值是byte_card，位置是int_pos可省略，如果位置不是(最大index+1)，原来位置的牌会被替换
---@param e_index UserIndex
---@param e_fromIndex UserIndex @用于显示箭头，如果是暗杠的，可以传nil，则不显示箭头
---@param e_kind HandleType
---@param byte_card int
---@param optional int_pos int
function cards_3d_layout:addHeapCard(e_index, e_fromIndex, e_kind, byte_card, int_pos)
    if int_pos == nil then
        int_pos = #self._heap_cards[e_index] + 1
    end
    if not self._heap_cards[e_index][int_pos] then
        local card_info,base_x,base_y,base_zorder = self:getLayoutInfo(e_index, "heap", "items", int_pos)
        -- 验证json里的"id"字段是否合法，目前看是顺序排列，从0开始的，如果将来有变化，需要改变逻辑找到正确的id对应的配置
        if card_info.id == int_pos - 1 then
            self._heap_cards[e_index][int_pos] = CHandledCards3D.new(card_info, base_x,base_y,base_zorder)
            for i = 1, 4 do
                self._view_root:addChild(self._heap_cards[e_index][int_pos]._handled_card[i])
            end
            self._view_root:addChild(self._heap_cards[e_index][int_pos])
            self._view_root:addChild(self._heap_cards[e_index][int_pos]._arrow)
        end
    end

    if self._heap_cards[e_index][int_pos] then
        self._heap_cards[e_index][int_pos]:setHandledData(e_fromIndex, e_kind, byte_card)
        self._heap_cards[e_index][int_pos]:showCards(self._corner_config, self._dark_card_value)
--        if self._card_mask[e_index] then
--            self._heap_cards[e_index][int_pos]:showMask(true);
--        end
    end
end

-- 从吃碰杠牌堆删除一组牌，位置是int_pos可省略
---@param e_index UserIndex
---@param optional int_pos int
function cards_3d_layout:removeHeapCard(e_index, int_pos)
    if self._heap_cards[e_index] and #self._heap_cards[e_index] > 0 then
        if not int_pos then
            int_pos = #self._heap_cards[e_index]
        end
        self._heap_cards[e_index][int_pos]:hideCards()
    end
end

-- 添加一张牌到花牌，牌值是byte_card，位置是int_pos可省略，如果位置不是(最大index+1)，原来位置的牌会被替换
---@param e_index UserIndex
---@param byte_card int
---@param optional int_pos int
function cards_3d_layout:addHuaCard(e_index, byte_card, int_pos)
    if int_pos == nil then
        int_pos = #self._hua_cards[e_index] + 1
    end
    if not self._hua_cards[e_index][int_pos] then
        local card_info,base_x,base_y,base_zorder = self:getLayoutInfo(e_index, "hua", "items", int_pos)
        -- 验证json里的"id"字段是否合法，目前看是顺序排列，从0开始的，如果将来有变化，需要改变逻辑找到正确的id对应的配置
        if card_info.id == int_pos - 1 then
            self._hua_cards[e_index][int_pos] = CTiledCard3D.new(card_info,base_x,base_y,base_zorder)
            self._view_root:addChild(self._hua_cards[e_index][int_pos])
        end
    end

    if self._hua_cards[e_index][int_pos] then
        self._hua_cards[e_index][int_pos]:showCard(byte_card, self._corner_config[byte_card], byte_card == self._dark_card_value)
    end
end

-- 从花牌删除一张牌，位置是int_pos可省略
---@param e_index UserIndex
---@param optional int_pos int
function cards_3d_layout:removeHuaCard(e_index, int_pos)
    if self._hua_cards[e_index] and #self._hua_cards[e_index] > 0 then
        if not int_pos then
            int_pos = #self._hua_cards[e_index]
        end
        self._hua_cards[e_index][int_pos]:hideCard()
    end
end

local index_table = {
    ["heap"] = {LayoutIndex.heap_bottm, LayoutIndex.heap_left, LayoutIndex.heap_right, LayoutIndex.heap_top},
    ["hua"] = {LayoutIndex.hua_bottm, LayoutIndex.hua_left, LayoutIndex.hua_right, LayoutIndex.hua_top},
    ["out"] = {LayoutIndex.out_bottm, LayoutIndex.out_left, LayoutIndex.out_right, LayoutIndex.out_top},
    ["show"] = {LayoutIndex.show_bottm, LayoutIndex.show_left, LayoutIndex.show_right, LayoutIndex.show_top},
    ["wall"] = {LayoutIndex.wall_bottm, LayoutIndex.wall_left, LayoutIndex.wall_right, LayoutIndex.wall_top},
}
-- 由UserIndex和"heap"/"out"等字符串获取LayoutIndex
---@param e_index UserIndex
---@param s_type string
---@return e_layout_index LayoutIndex
function cards_3d_layout:getLayoutIndex(e_index, s_type)
    local ret_tbl = index_table[s_type]
    if ret_tbl == nil then return nil end
    if e_index == UserIndex.down then
        return ret_tbl[1]
    elseif e_index == UserIndex.left then
        return ret_tbl[2]
    elseif e_index == UserIndex.right then
        return ret_tbl[3]
    elseif e_index == UserIndex.up then
        return ret_tbl[4]
    end
    return nil
end

-- e_index是UserIndex，s_type是"heap"等字符串，s_inner_type是"items"或"items_5"等内部序号（带数字的只在手牌数据中有），int_pos是序号
---@param e_index UserIndex
---@param s_type string
---@param s_inner_type string
---@param int_pos int
---@return card_info table
---@return base_x int
---@return base_y int
---@return base_zorder int
function cards_3d_layout:getLayoutInfo(e_index, s_type, s_inner_type, int_pos)
    local i = self:getLayoutIndex(e_index, s_type)
    local layout = self._layout[i]
    local base_zorder = layout.zorder
    local base_x
    local base_y
    if self._layout_offset[i].forceX then
        base_x = self._layout_offset[i].x
    else
        base_x = layout.x
    end
    if self._layout_offset[i].forceY then
        base_y = self._layout_offset[i].y
    else
        base_y = layout.y
    end

    if write_online_logs then
        if i <= LayoutIndex.heap_top then
            if int_pos > 5 or int_pos < 1 then
                write_online_logs("i:"..tostring(i).." pos:"..tostring(int_pos));
            end
        elseif i <= LayoutIndex.hua_top then
            if int_pos > 15 or int_pos < 1  then
                write_online_logs("i:"..tostring(i).." pos:"..tostring(int_pos));
            end
        elseif i <= LayoutIndex.out_top then    
            if int_pos > 36 or int_pos < 1 then
                write_online_logs("i:"..tostring(i).." pos:"..tostring(int_pos));
            end
        elseif i <= LayoutIndex.wall_top then 
            if int_pos > 17 or int_pos < 1 then
                write_online_logs("i:"..tostring(i).." pos:"..tostring(int_pos));
            end
        end
    end
    
    local card_info = layout[s_inner_type][int_pos]
    return card_info, base_x, base_y, base_zorder
end

-- 从json文件读取坐标等信息
---@param s_path string
---@param e_index LayoutIndex
function cards_3d_layout:readLayoutFromFile(s_path, e_index)
    self._layout[e_index] = game_tools.decode(io.readfile(s_path))
end

-- 初始化坐标
---@param s_path string
---@param view_size table
function cards_3d_layout:initLayout(s_path, view_size)
    -- 读取所有坐标文件
    self:readLayoutFromFile(s_path.."layout_heap_cards_bottom.json",    LayoutIndex.heap_bottm)
    self:readLayoutFromFile(s_path.."layout_heap_cards_left.json",      LayoutIndex.heap_left)
    self:readLayoutFromFile(s_path.."layout_heap_cards_right.json",     LayoutIndex.heap_right)
    self:readLayoutFromFile(s_path.."layout_heap_cards_top.json",       LayoutIndex.heap_top)

    self:readLayoutFromFile(s_path.."layout_hua_cards_bottom.json",     LayoutIndex.hua_bottm)
    self:readLayoutFromFile(s_path.."layout_hua_cards_left.json",       LayoutIndex.hua_left)
    self:readLayoutFromFile(s_path.."layout_hua_cards_right.json",      LayoutIndex.hua_right)
    self:readLayoutFromFile(s_path.."layout_hua_cards_top.json",        LayoutIndex.hua_top)

    self:readLayoutFromFile(s_path.."layout_out_cards_bottom.json",     LayoutIndex.out_bottm)
    self:readLayoutFromFile(s_path.."layout_out_cards_left.json",       LayoutIndex.out_left)
    self:readLayoutFromFile(s_path.."layout_out_cards_right.json",      LayoutIndex.out_right)
    self:readLayoutFromFile(s_path.."layout_out_cards_top.json",        LayoutIndex.out_top)

    self:readLayoutFromFile(s_path.."layout_show_cards_bottom.json",    LayoutIndex.show_bottm)
    self:readLayoutFromFile(s_path.."layout_show_cards_left.json",      LayoutIndex.show_left)
    self:readLayoutFromFile(s_path.."layout_show_cards_right.json",     LayoutIndex.show_right)
    self:readLayoutFromFile(s_path.."layout_show_cards_top.json",       LayoutIndex.show_top)

    self:readLayoutFromFile(s_path.."layout_wall_cards_bottom.json",    LayoutIndex.wall_bottm)
    self:readLayoutFromFile(s_path.."layout_wall_cards_left.json",      LayoutIndex.wall_left)
    self:readLayoutFromFile(s_path.."layout_wall_cards_right.json",     LayoutIndex.wall_right)
    self:readLayoutFromFile(s_path.."layout_wall_cards_top.json",       LayoutIndex.wall_top)

    self:configLayoutOffset(view_size)

    -- 其他一些坐标初始化
    self._handle_ani_pos={
        ccp(240,view_size.height*0.5+150), -- left
        ccp(view_size.width*0.5,(view_size.width+800)/view_size.height*(300/((1000+800)/750))), -- down
        ccp(view_size.width-240,view_size.height*0.5+150), -- right
        ccp(view_size.width*0.5,view_size.height-100) -- up
    }
    if view_size.width/view_size.height > 17/9 then
        self._handle_ani_pos={
            ccp(300,view_size.height*0.5+150), -- left
            ccp(view_size.width*0.5,(view_size.width+800)/view_size.height*(300/((1000+800)/750))), -- down
            ccp(view_size.width-300,view_size.height*0.5+150), -- right
            ccp(view_size.width*0.5,view_size.height-100) -- up
        }
    end
end

-- 根据屏幕适配各个锚点偏移
---@param view_size table
function cards_3d_layout:configLayoutOffset(view_size)
    -- 根据桌子两侧贴图的位置来决定牌堆锚点，贴图位置会随着分辨率不同而变化，中间部分最小不小于640像素
    local left_anchor_x = self._table_3d_layout._left_table_bg:getPositionX()
    local right_anchor_x = self._table_3d_layout._right_table_bg:getPositionX()
    -- 位置原本是屏幕中心原点
    left_anchor_x = view_size.width/2 + left_anchor_x
    right_anchor_x = view_size.width/2 + right_anchor_x
    -- 牌堆锚点距离贴图位置的偏移
    local x_offset = 40
    left_anchor_x = left_anchor_x + x_offset
    right_anchor_x = right_anchor_x - x_offset

    local left_out_card_gap = view_size.width/2 - 377.5 + 40;
    local right_out_card_gap = view_size.width/2 + 377.5 -40;
--    print("letf:"..left_anchor_x.." right:"..right_anchor_x);
--    print("letf:"..left_out_card_gap.." right:"..right_out_card_gap);

    self._layout_offset = {}
    self._layout_offset[LayoutIndex.heap_top]   = {forceX=true, forceY=false, x=view_size.width/2, y=0}
    self._layout_offset[LayoutIndex.hua_top]    = {forceX=true, forceY=false, x=view_size.width/2, y=0}
    self._layout_offset[LayoutIndex.show_top]   = {forceX=true, forceY=false, x=view_size.width/2, y=0}
    self._layout_offset[LayoutIndex.wall_top]   = {forceX=true, forceY=false, x=view_size.width/2, y=0}
    self._layout_offset[LayoutIndex.out_top]    = {forceX=true, forceY=false, x=view_size.width/2, y=0}

    self._layout_offset[LayoutIndex.heap_bottm] = {forceX=false, forceY=false, x=0, y=0}
    self._layout_offset[LayoutIndex.hua_bottm]  = {forceX=true, forceY=false, x=view_size.width/2, y=0}
    self._layout_offset[LayoutIndex.show_bottm] = {forceX=false, forceY=false, x=0, y=0}
    self._layout_offset[LayoutIndex.wall_bottm] = {forceX=false, forceY=false, x=0, y=0}
    self._layout_offset[LayoutIndex.out_bottm]  = {forceX=true, forceY=false, x=view_size.width/2, y=0}

    self._layout_offset[LayoutIndex.heap_left]  = {forceX=true, forceY=false, x=left_anchor_x, y=0}
    self._layout_offset[LayoutIndex.hua_left]   = {forceX=true, forceY=false, x=left_anchor_x+60, y=0}
    self._layout_offset[LayoutIndex.show_left]  = {forceX=true, forceY=false, x=left_anchor_x, y=0}
    self._layout_offset[LayoutIndex.wall_left]  = {forceX=true, forceY=false, x=left_anchor_x, y=0}
    self._layout_offset[LayoutIndex.out_left]   = {forceX=true, forceY=false, x=left_out_card_gap, y=10}

    self._layout_offset[LayoutIndex.heap_right] = {forceX=true, forceY=false, x=right_anchor_x, y=0}
    self._layout_offset[LayoutIndex.hua_right]  = {forceX=true, forceY=false, x=right_anchor_x-60, y=0}
    self._layout_offset[LayoutIndex.show_right] = {forceX=true, forceY=false, x=right_anchor_x, y=0}
    self._layout_offset[LayoutIndex.wall_right] = {forceX=true, forceY=false, x=right_anchor_x, y=0}
    self._layout_offset[LayoutIndex.out_right]  = {forceX=true, forceY=false, x=right_out_card_gap, y=10}
end

return cards_3d_layout
