require ("src/cards_3d_config")
---@class cards_3d_logic @框架牌逻辑层
cards_3d_logic=class("cards_3d_logic")

-- 公共接口
function cards_3d_logic:ctor(...)
    cards_3d_logic:init(...)
end

-- 初始化，进游戏先调一次
---@param view_root table
function cards_3d_logic:init()
    self:reset()
end

-- 清空所有牌并重置状态，每盘游戏开局调一次！
function cards_3d_logic:reset()
    self._cards = {
        {{},{},{},{}},
        {{},{},{},{}},
        {{},{},{},{}},
        {{},{},{},{}}
    }
end

-- 释放资源，游戏退出时调用
function cards_3d_logic:destroy()
    self:reset()
end

-- 某个方位抓牌
---@param e_index UserIndex
---@param byte_card int
function cards_3d_logic:catchCard(e_index, byte_card)
    self:addCard(e_index, LayoutCardsType.hand, byte_card)
    --self._last_op_card_index = e_index
end

-- 某个方位出牌
---@param e_index UserIndex
---@param byte_card int
function cards_3d_logic:outCard(e_index, byte_card)
    self:removeCardByValue(e_index, LayoutCardsType.hand, byte_card)
    self:addCard(e_index, LayoutCardsType.out, byte_card)
    --self._last_op_card_index = e_index
end

-- 某个方位从另一个方位的玩家吃/碰/杠操作
---@param e_index UserIndex
---@param e_fromIndex UserIndex @考虑到断线重连，必须指定from id，如果是暗杠，可以传nil，则不显示箭头
---@param e_heapType HandleType
---@param byte_value int
function cards_3d_logic:handle(e_index, e_fromIndex, e_heapType, byte_value)
    if e_heapType == HandleType.b_chi_left then
        self:removeCard(e_fromIndex, LayoutCardsType.out)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value+1)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value+2)
        self:addHeap(e_index, e_fromIndex, e_heapType, byte_value)
    elseif e_heapType == HandleType.b_chi_mid then
        self:removeCard(e_fromIndex, LayoutCardsType.out)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value-1)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value+1)
        self:addHeap(e_index, e_fromIndex, e_heapType, byte_value)
    elseif e_heapType == HandleType.b_chi_right then
        self:removeCard(e_fromIndex, LayoutCardsType.out)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value-2)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value-1)
        self:addHeap(e_index, e_fromIndex, e_heapType, byte_value)
    elseif e_heapType == HandleType.b_peng then
        self:removeCard(e_fromIndex, LayoutCardsType.out)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:addHeap(e_index, e_fromIndex, e_heapType, byte_value)
    elseif e_heapType == HandleType.b_gang then
        self:removeCard(e_fromIndex, LayoutCardsType.out)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:addHeap(e_index, e_fromIndex, e_heapType, byte_value)
    elseif e_heapType == HandleType.d_gang then
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        self:addHeap(e_index, e_fromIndex, e_heapType, byte_value)
    elseif e_heapType == HandleType.a_gang then
        self:removeCardByValue(e_index, LayoutCardsType.hand, byte_value)
        for i=1,#self._cards[LayoutCardsType.heap][e_index] do
            local heap = self._cards[LayoutCardsType.heap][e_index][i]
            if heap.type == HandleType.b_peng and heap.value == byte_value then
                heap.type = HandleType.a_gang
                break
            end
        end
    else
        -- error
    end
end

-- 获取某个方位某个队列的长度
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@return n int
function cards_3d_logic:getCardsLength(e_index, e_type)
    return #self._cards[e_type][e_index]
end

-- 某个方位补花
---@param e_index UserIndex
---@param byte_card int
function cards_3d_logic:buHua(e_index, byte_card)
    self:removeCardByValue(e_index, LayoutCardsType.hand, byte_card)
    self:addCard(e_index, LayoutCardsType.hua, byte_card)
end

---- 给某个方位某个队列排序
-----@param e_index UserIndex
-----@param e_type LayoutCardsType
--function cards_3d_logic:sortCard(e_index, e_type)
--    if e_type == LayoutCardsType.heap then return end
--    table.sort(self._cards[e_type][e_index])
--end

-- 获取某个方位某个队列里的牌数组
---@param e_index UserIndex
---@param e_type LayoutCardsType
function cards_3d_logic:getCards(e_index, e_type)
    return self._cards[e_type][e_index]
end

-- 移动某个方位某个队列里的牌，注意这里只是逻辑上的顺序
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@param int_pos_from int @被移动的牌的index
---@param int_pos_to int @要移动到的index
function cards_3d_logic:moveCard(e_index, e_type, int_pos_from, int_pos_to)
    if int_pos_from < int_pos_to then
        -- 从左往右移，先添加再删除
        local c = self._cards[e_type][e_index][int_pos_from]
        table.insert(self._cards[e_type][e_index], int_pos_to, c)
        table.remove(self._cards[e_type][e_index], int_pos_from)
    elseif int_pos_from > int_pos_to then
        -- 从右往左移，先删除再添加
        local c = self._cards[e_type][e_index][int_pos_from]
        table.remove(self._cards[e_type][e_index], int_pos_from)
        table.insert(self._cards[e_type][e_index], int_pos_to, c)
    else
        -- 不变
    end
end

function cards_3d_logic:setSortCardsFunc(func_sortCard)
    if func_sortCard and type(func_sortCard) == "function" then
        self._sort_cards_func=func_sortCard
    end
end

function cards_3d_logic:sortHandCards(e_index)
    if self._sort_cards_func and e_index then
        local cards = self._sort_cards_func(self._cards[LayoutCardsType.hand][e_index])
        local list_cpy = clone(cards)
        cards_3d_logic:setCards(e_index, LayoutCardsType.hand, list_cpy)
    end
end

-- 以下原则上是private方法

-- 设置某个方位某个队列里的所有牌
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@param cards table {11,12,13...}
function cards_3d_logic:setCards(e_index, e_type, cards)
    if e_type == LayoutCardsType.heap then return end
    if cards == nil then
        print("error cards");
    end
    self._cards[e_type][e_index] = cards
end

-- 往某个方位某个队列里加入一张牌
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@param byte_card int
---@param optional int_pos int @插入位置，如果不传，则插入到最后
function cards_3d_logic:addCard(e_index, e_type, byte_card, int_pos)
    if e_type == LayoutCardsType.heap then return end
    if int_pos then
        table.insert(self._cards[e_type][e_index], int_pos, byte_card)
    else
        table.insert(self._cards[e_type][e_index], byte_card)
    end
end

-- 往某个方位某个队列里移除一张牌，如果不指定位置，移除最后一张
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@param optional pos int
function cards_3d_logic:removeCard(e_index, e_type, pos)
    if e_type == LayoutCardsType.heap then return end
    if not pos then pos = #self._cards[e_type][e_index] end
    table.remove(self._cards[e_type][e_index], pos)
end

-- 往某个方位某个队列里移除一张指定牌值的牌
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@param byte_card int
function cards_3d_logic:removeCardByValue(e_index, e_type, byte_card)
    if e_type == LayoutCardsType.heap then return end

    -- 必须要从后往前搜索，如果玩家出的是最后一张，不会触发插牌操作，也不会触发逻辑上的重新排序，所以要确保出掉之后手牌顺序和原来一样
    for i=#self._cards[e_type][e_index],1,-1 do
        if self._cards[e_type][e_index][i] == byte_card then
            table.remove(self._cards[e_type][e_index], i)
            return
        end
    end

    -- 没有找到，移除最后一张，在移除其余玩家的手牌时会调用到，保持接口统一
    if e_type == LayoutCardsType.hand then
        self:removeCard(e_index, LayoutCardsType.hand)
    end
end

-- 设置某个方位吃碰杠牌堆
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@param heaps table @{{type=HandleType.b_peng, value=21, from=UserIndex.up},{type=HandleType.b_gang, value=24, from=UserIndex.left}}
function cards_3d_logic:setHeaps(e_index, heaps)
    self._cards[LayoutCardsType.heap][e_index] = heaps
end

-- 往某个方位吃碰杠牌堆里加入一组牌
---@param e_index UserIndex
---@param e_heapType HandleType
---@param byte_value int
function cards_3d_logic:addHeap(e_index, e_fromIndex, e_heapType, byte_value)
    --{type=HandleType.b_peng, value=21}
    table.insert(self._cards[LayoutCardsType.heap][e_index], {type=e_heapType, value=byte_value, from=e_fromIndex})
end

-- 往某个方位吃碰杠牌堆里移除一组牌，如果不指定位置，移除最后一组
---@param e_index UserIndex
---@param e_type LayoutCardsType
---@param optional pos int
function cards_3d_logic:removeHeap(e_index, pos)
    if not pos then pos = #self._cards[LayoutCardsType.heap][e_index] end
    table.remove(self._cards[LayoutCardsType.heap][e_index], pos)
end

return cards_3d_logic
