--------------------------------------------------------------------------------
-- 玩家头像
--------------------------------------------------------------------------------
game_user = class("game_user")

function game_user.extend(target)
	local t = tolua.getpeer(target)
	if not t then
		t = {}
		tolua.setpeer(target, t)
	end
	setmetatable(t, game_user)
	return target
end

--------------------------------------------------------------------------------
-- 常量
--------------------------------------------------------------------------------
local tag = {self = 1, enemy = 2, needless = 3, base = 4}

local ccs = ccs or {}
ccs.MovementEventType = {
    START = 0,
    COMPLETE = 1,
    LOOP_COMPLETE = 2, 
}

--------------------------------------------------------------------------------
-- 属性
--------------------------------------------------------------------------------
game_user.player_background = nil
game_user.player_heads = nil
game_user.player_status = nil 
game_user.player_trust = nil
game_user.player_identity = nil
game_user.player_nickname = nil
--game_user.player_gold = nil
game_user.cards_count_bg = nil
game_user.cards_count = nil
game_user.player_pos = nil
game_user.battery_num = -1
game_user.player_vip = nil
--game_user.gold_img = nil

game_user.status = nil
--------------------------------------------------------------------------------
-- 函数
--------------------------------------------------------------------------------

function game_user:clear_user_data()
    for i=1, GAME_PLAYER do
        self:set_identity(i, -1)
--        self:set_status(i, UserStatus.null)
    end
end

function game_user:turn_outline_status(viewid, isOnline)
    if isOnline == true then
        self.player_background[viewid]:loadTexture("head_bg.png",1)
        self.outline_img[viewid]:setVisible(false)
    else
        self.player_background[viewid]:loadTexture("head_bg_1.png",1)
        self.outline_img[viewid]:setVisible(true)
    end
end

function game_user:set_head_bg(viewid, head_type)
    if not viewid or viewid < 0 or viewid > GAME_PLAYER then
        return 
    end
    
    if head_type == 1 then
        self.player_background[viewid]:loadTexture("head_bg_leave.png",1);
    else 
        self.player_background[viewid]:loadTexture("head_bg.png",1);
    end
    
end

function game_user:set_identity(viewid, identity)
    if not viewid or viewid < 0 or viewid > GAME_PLAYER then
        return 
    end

    if self.player_background[viewid]:isVisible() == false then
        return
    end

    if identity == -1 then
        self.player_identity[viewid]:setVisible(false)
        return 
    end

    self.player_identity[viewid]:setVisible(true)
end

function game_user:get_status(viewid)
    if not viewid or viewid < 0 or viewid > GAME_PLAYER then
        return 0;
    end

    return self.status[viewid];
end

function game_user:set_status(viewid, status)
    if not viewid or viewid < 0 or viewid > GAME_PLAYER then
        return 
    end

    self.status[viewid] = status;

    if #self.player_status < GAME_PLAYER then
        return;
    end

    if status ~= UserStatus.ready then
        self.player_status[viewid]:setVisible(false)
        --UIHelper.show_clock(viewid, 0)        
        return
    end

    self.player_status[viewid]:setVisible(true)
end

function game_user:set_sex(viewid, userid, faceid, sex)
    if not viewid or viewid < 0 or viewid > GAME_PLAYER then
        return 
    end

    self.player_heads[viewid]:__set_userid(userid);

--    if self.player_heads[viewid] == nil then
--        return 
--    end
--    if self.player_heads[viewid].ui_head == nil then
--        return 
--    end
--    self.player_heads[viewid].ui_head:set_head(80, 80, userid, faceid, sex)

    game_layer.table_user_sex[viewid]=sex;
    game_layer.table_user_userid[viewid]=userid
    game_layer.table_user_faceid[viewid]=faceid
end

function game_user:show_user(viewid, isShow, isAction, isNocheat)
    isAction = isAction ~= nil and isAction or false
    isNocheat = isNocheat ~= nil and isNocheat or false

    self.player_background[viewid]:setVisible(isShow)

    if isShow == false and viewid ~= LocalSelfChairId and isAction then
        local animation = CCAnimation:create()
        for i=1, 7 do
--            local sprit_frame = CCSpriteFrame:create(("majiang/leave_action/leave" .. i .. ".png"),CCRectMake(0, 0, 64, 84))
            local sprit_frame = CCSpriteFrameCache:sharedSpriteFrameCache():spriteFrameByName("leave"..i..".png")
            animation:addSpriteFrame(sprit_frame)
        end
        animation:setDelayPerUnit(0.7 / 7)
        animation:setRestoreOriginalFrame(true)
        local action = CCAnimate:create(animation)
        self.player_sprite[viewid]:setPosition(ccp(self.player_background[viewid]:getPositionX(), self.player_background[viewid]:getPositionY() + 13))
        self.player_sprite[viewid]:runAction(action)
    end

    if game_layer.isCanChangeSeat ~= 0 then
        self.changeSeat[viewid]:setVisible(not isShow);
    end
end

function game_user:resetChangeSeat()
    for i=1,GAME_PLAYER do 
        if not self.player_background[i]:isVisible() then
            self.changeSeat[i]:setVisible(true);
        end
    end
  
end

function game_user:init( visibleSize ) 
    self:setSize(visibleSize)

    self.player_background = {}
    self.player_heads = {}
    self.player_gold = {}
    self.player_identity = {}
    self.player_leave_img = {}
    self.player_status = {}
    self.outline_img = {}
    self.player_sprite = {}
    self.player_name = {}
    self.player_voice_armature = {}
    self.player_clock = {}
    self.player_hua_bg = {}
    self.player_hua_label = {}
    self.player_vip = {}
    self.status = {}

    self.changeSeat = {}

    self.player_pos_1 = {ccp(visibleSize.width/2, visibleSize.height - 100), ccp(visibleSize.width - 200, 380), ccp(visibleSize.width/2, 160), ccp(200, 380)}
    self.player_pos = {ccp(visibleSize.width * 0.84, visibleSize.height - 50), ccp(visibleSize.width - 46 - game_layer.screenGap, 380), ccp(46 + game_layer.screenGap, 115), ccp(46 + game_layer.screenGap, 380)}

    for i=1, VIEW_INDEX_COUNT do
        -- 头像背景
        local head_back = UIHelper.create_img(self, self.player_pos_1[i], "head_bg.png",1)
        head_back:setVisible(false)
        table.insert(self.player_background, head_back)

        --换座
        local btnChangeSeat = UIHelper.create_btn(self, self.player_pos_1[i], "changeSeat.png", 1);
        table.insert(self.changeSeat, btnChangeSeat);
        btnChangeSeat:setVisible(false);
        --换座点击事件
        local function OnEventBtnChangeSeat(sender, touchType)
            if touchType ~= _G.TOUCH_EVENT_ENDED then
                return;
            end
            local seatId = UIHelper.switch_to_seat_id(i);
	        local tbl = {}
            tbl.seat = seatId
            frame_tools:send_data(MessageType.SUB_C_CHANGE_SEAT, tbl)
        end
        btnChangeSeat:addTouchEventListener(OnEventBtnChangeSeat);

        -- 头像
        local btn_head = frame_tools:create_head_item() -- UIHelper.game_user_head(head_back, ccp(0, 13))
        head_back:addChild(btn_head);
        btn_head:setPosition(ccp(0,13))
        btn_head:__set_head_scale(86/120);
        table.insert(self.player_heads, btn_head)

        local gold_control = btn_head.__gold_bg;
        gold_control:setPosition(ccp(-5, -65));

        btn_head.__head_fg:setVisible(false);

        --名字
        local name_label = btn_head.__name_label;
        table.insert(self.player_name, name_label)
        name_label:setPosition(ccp(0, -10));

        --花
        local img_hua = UIHelper.create_img(head_back,ccp(0,-70),"hua_bg.png",1);
        table.insert(self.player_hua_bg,img_hua);
        img_hua:setVisible(false);
        img_hua:setScale(0.8);

        local label_hua = UIHelper.create_labelBM(img_hua,"resources/number_gold/number_fuzhoumajiang_hua.fnt",ccp(10,-10))
        table.insert(self.player_hua_label,label_hua);

        --计时器
        local local_clock = user_color:create(CCSizeMake(80,80),i);
        btn_head:addChild(local_clock);
        table.insert(self.player_clock,local_clock);  
        
        self.have_move=false;
        -- 庄
        local banker_img = nil
        if i == 2 or i == 1 then
            banker_img = UIHelper.create_img(head_back, ccp(-40, 45), "banker.png",1)
        else
            banker_img = UIHelper.create_img(head_back, ccp(40, 45), "banker.png",1)
        end
        banker_img:setVisible(false)
        table.insert(self.player_identity, banker_img)

        --离线
        local outline_img = UIHelper.create_img(btn_head, ccp(-head_back:getContentSize().width/2 + 25, -head_back:getContentSize().height/2 + 28), "out_line.png",1)
        outline_img:setVisible(false)
        table.insert(self.outline_img, outline_img)

        -- 离开动画
        local sprite = CCSprite:create()
        sprite:setPosition(ccp(self.player_pos_1[i].x, self.player_pos_1[i].y + 13))
        self:addNode(sprite)
        table.insert(self.player_sprite, sprite)

        -- 状态(是否已准备)
        table.insert(self.status,0);
        local status = CCArmature:create("dh_zbz")
        if status then
            if i == 1 then
                status:setPosition(ccp(self.player_pos_1[i].x, self.player_pos_1[i].y - 100 ))
            elseif i == LocalSelfChairId then
                status:setPosition(ccp(self.player_pos_1[i].x - 130, self.player_pos_1[i].y))
            elseif i == LocalSelfChairId + 1 then
                status:setPosition(ccp(game_layer.visibleSize.width/2, self.player_pos_1[i].y + 80))
            else
                status:setPosition(ccp(self.player_pos_1[i].x + 100, self.player_pos_1[i].y))
            end
            status:getAnimation():playWithIndex(0)
            status:setVisible(false)
            self:addNode(status)
            table.insert(self.player_status, status)
        end

    end
end

function game_user:move_head(move_type)
    
    if self.have_move then
        return;
    end
    for k,v in pairs(self.player_background) do
        self.player_hua_label[k]:setText("x0");

        local array =CCArray:create()
        array:addObject(CCMoveTo:create(0.5, self.player_pos[k]))
        array:addObject(CCCallFunc:create(function()
		game_layer.init_chat_layer(1)
	    end))
        array:addObject(CCCallFunc:create(function()
        self.player_hua_bg[k]:setOpacity(0);
        self.player_hua_label[k]:setOpacity(0);

		self.player_hua_bg[k]:setVisible(true);

        self.player_hua_bg[k]:runAction(CCFadeIn:create(2))
        self.player_hua_label[k]:runAction(CCFadeIn:create(2))

        self.changeSeat[k]:setPosition(self.player_pos[k]);
	    end))
        v:runAction(CCSequence:create(array))
        v:runAction(CCScaleTo:create(0.5,0.75))
        self.player_name[k]:setVisible(false);
    end
    self.have_move = true;

end

function game_user:set_name(int_view,name)
    if not int_view or int_view < 0 or int_view > GAME_PLAYER then
        return 
    end
--    self.player_name[int_view]:setText(name);
end

function game_user:set_clock(int_view, int_time, all_time, bool_exit)
    if not int_view or int_view < 0 or int_view > GAME_PLAYER then
        return 
    end

    self.player_clock[int_view]:set_user_color(int_time,all_time,bool_exit);
end

function game_user:set_hua_count(int_view,count)
    if not int_view or int_view < 0 or int_view > GAME_PLAYER or count == nil then
        return 
    end

    self.player_hua_bg[int_view]:setVisible(true);
    self.player_hua_label[int_view]:setText("x"..count);
end

function game_user:show_hua_count(bool_show)
    bool_show = bool_show or false;
    if bool_show then
        for i=1,GAME_PLAYER,1 do
            self.player_hua_bg[i]:setVisible(true);
            self.player_hua_label[i]:setText("x0");
        end
    else 
        for i=1,GAME_PLAYER,1 do
            self.player_hua_bg[i]:setVisible(false);
        end
    end
end

function game_user:set_head_able(int_view, args)
    if not int_view or int_view < 0 or int_view > GAME_PLAYER then
        return 
    end

    self.player_heads[int_view]:setTouchEnabled(args);
end

function game_user:reset(args)
    game_layer.game_user:show_hua_count(false)
    game_layer.game_user:set_clock(VIEW_INDEX_SELF,-1,-1);

    self.have_move = false;
    self:clear_user_data();
    for k,v in pairs(self.player_background) do
        v:setPosition(self.player_pos_1[k]);
        v:setScale(1);
        self.player_name[k]:setVisible(true);

        self.changeSeat[k]:setPosition(self.player_pos_1[k]);
    end
    game_layer.init_chat_layer()
end

function game_user:create(visibleSize)
	local layer = game_user.extend(Layout:create())
    if layer ~= nil then
        layer:init(visibleSize)
    end
	return layer
end


--------------------------------------------------------------------------------
-- 准备计时
--------------------------------------------------------------------------------

user_color = class("user_color")

--------------------------------------------------------------------------------
-- 参数
--------------------------------------------------------------------------------
user_color.view = -1;
user_color.fTimer=0;
user_color.int_time=-1;
user_color.all_time=-1;
user_color.int_id=-1;

user_color.foreground=nil
--------------------------------------------------------------------------------
-- 函数
--------------------------------------------------------------------------------
function user_color.extend(target)
    local t = tolua.getpeer(target)
    if not t then
        t = {}
        tolua.setpeer(target, t)
    end
    setmetatable(t, user_color)
    return target
end

function user_color:init(the_size,view)
    self.foreground = UIHelper.create_progress(self,ccp(0,0),"finish.png",1,_G.kCCProgressTimerTypeRadial,true)
    self.foreground:setVisible(false)

    self.view = view;
end

function user_color:set_user_color(int_time,all_time,bool_exit)
    if int_time == nil or all_time == nil then
        return;
    end
    if all_time <= 0 then
        self.foreground:setVisible(false)
        self:unscheduleUpdate()
        return;
    end
    if int_time>0 then
        self.foreground:setVisible(true)
        self.all_time = all_time;
        self.int_time = int_time;
        self.foreground:setPercentage(100)
        local function update(dt)
            if self.int_time<=0 then
                return;
            end

            self.int_time = self.int_time - dt
            if self.int_time > 0 then
                self.foreground:setPercentage(self.int_time/self.all_time * 100)
            elseif self.int_time <= 0 then
                self:unscheduleUpdate()
                if self.view == VIEW_INDEX_SELF and bool_exit then
                    game_event.on_start_time_over()
                end
            end
        end
        self:scheduleUpdateWithPriorityLua(update, 0)
    else
        self.foreground:setVisible(false)
        self:unscheduleUpdate()
    end
end

function user_color:create(the_size,view)
    local color = user_color.extend(ImageView:create())
    color:init(the_size,view)
    return color
end
