--------------------------------------------------------------------------------
-- 礼物发送
--------------------------------------------------------------------------------
gift_play = class("gift_play")

function gift_play.extend(target)
	local t = tolua.getpeer(target)
	if not t then
		t = {}
		tolua.setpeer(target, t)
	end
	setmetatable(t, gift_play)
	return target
end

--------------------------------------------------------------------------------
-- 函数
--------------------------------------------------------------------------------
function gift_play:play_gift(from_pos, to_pos, gift_id, sprite_id, delay, scale_info)
    sprite_id = sprite_id or 0
    delay = delay or 0

    if gift_id < 0 or gift_id > 4 then
        return
    end

    local file_info = CCFileUtils:sharedFileUtils():fullPathForFilename("gift/"..gift_id.."/config.json")
    if file_info == nil then
        return 
    end
    
    local str_info = CCString:createWithContentsOfFile(file_info):getCString()
    local c = require "cjson"
    local table_json = c.decode(str_info)
    if table_json == nil then
        return 
    end

    local start_width = table_json.start_width
    local start_height = table_json.start_height
    local start_count = table_json.start_count
    local start_time = table_json.start_time
    local animation1 = CCAnimation:create()
    for i=1, start_count do
        local sprit_frame = CCSpriteFrame:create(("gift/" .. gift_id .. "/start" .. i .. ".png"),
                                            CCRectMake(0, 0, start_width, start_height))
        animation1:addSpriteFrame(sprit_frame)
    end
    animation1:setDelayPerUnit(start_time / start_count)
    animation1:setRestoreOriginalFrame(true)
    local start_action = CCAnimate:create(animation1)

    local move_width = table_json.move_width
    local move_height = table_json.move_height
    local move_time = table_json.move_time
    local move_rotation = table_json.move_rotation
    local move_turn = table_json.move_turn
    local animation2 = CCAnimation:create()
    local sprit_frame = CCSpriteFrame:create(("gift/" .. gift_id .. "/duration.png"),
                                        CCRectMake(0, 0, move_width, move_height))
    animation2:addSpriteFrame(sprit_frame)
    animation2:setDelayPerUnit(move_time / 1)
    animation2:setRestoreOriginalFrame(true)
    local move_action = CCAnimate:create(animation2)

    local end_width = table_json.end_width
    local end_height = table_json.end_height
    local end_count = table_json.end_count
    local end_time = table_json.end_time
    local animation3 = CCAnimation:create()
    for i=1, end_count do
        local sprit_frame = CCSpriteFrame:create(("gift/" .. gift_id .. "/end" .. i .. ".png"),
                                            CCRectMake(0, 0, end_width, end_height))
        animation3:addSpriteFrame(sprit_frame)
    end
    animation3:setDelayPerUnit(end_time / end_count)
    animation3:setRestoreOriginalFrame(true)

    local end_action = CCAnimate:create(animation3)

    local sprite_gift = {}
    sprite_gift[sprite_id] = CCSprite:create()
    sprite_gift[sprite_id]:setPosition(from_pos)
    self:addNode(sprite_gift[sprite_id])

    local function action_sound()
        audio_engine.playGiftEffect(gift_id)
    end

    local function callBack()
        sprite_gift[sprite_id]:removeFromParentAndCleanup(true)
    end

    local array = CCArray:create()
    array:addObject(CCDelayTime:create(delay))
    array:addObject(CCCallFunc:create(action_sound))
    array:addObject(move_action)
    array:addObject(end_action)
    array:addObject(CCCallFunc:create(callBack))
    sprite_gift[sprite_id]:runAction(CCSequence:create(array))

    local delay1 = CCDelayTime:create(delay)
    local move1 = CCMoveTo:create(move_time,to_pos)

    local array = CCArray:create()
    array:addObject(delay1)
    array:addObject(move1)
    sprite_gift[sprite_id]:runAction(CCSequence:create(array))
    
    if move_rotation == 1 then 
        local delay_time = CCDelayTime:create(delay)
        local rotate = CCRotateBy:create(move_time,360*move_turn)

        local array = CCArray:create()
        array:addObject(delay_time)
        array:addObject(rotate)
        sprite_gift[sprite_id]:runAction(CCSequence:create(array))
    end
    sprite_gift[sprite_id]:setScale(0.8 * scale_info)
end

function gift_play:play_gift_times(form_pos, to_pos, gift_id, gift_num, scale_info)
    gift_num = gift_num or 1
    if scale_info == nil then
        scale_info = 1
    end
    for i=1, gift_num do
        self:play_gift(form_pos, to_pos, gift_id, i, 0.2*(i-1), scale_info)
    end
end

function gift_play:create(visibleSize)
	local layer = gift_play.extend(Layout:create())
	layer:setSize(visibleSize)
	return layer
end