require ("src/cards_3d_config")
---@class table_3d_layout
table_3d_layout=class("table_3d_layout")

function table_3d_layout:ctor(...)
    self._view_size = nil
    self._view = nil
    self._bg = nil
    self._table = nil
    self._left_table_bg = nil
    self._right_table_bg = nil
	self._right_table_mask = nil
	self._left_table_mask = nil
    self._clock = nil
    self:init(...)
end

-- 公共方法

-- 初始化
---@param view_root table
---@param view_size table
function table_3d_layout:init(view_root, view_size)
    self._view_size = view_size
    self._view = control_tools.newImg({})
    view_root:addChild(self._view)

    CCSpriteFrameCache:sharedSpriteFrameCache():addSpriteFramesWithFile(Layout3DpListPath.."table.plist", Layout3DpListPath.."table.png")

    -- bg
    self._bg = control_tools.newImg({path=Layout3DTablePath.."room_bg1.png"})
    self._view:addChild(self._bg)
    self._bg:setScale(view_size.height/750);

    -- table root
    self._table_old = control_tools.newImg({})
    self._view:addChild(self._table_old)

    self._table = control_tools.newImg({path=Layout3DTablePath.."table.png"})
    self._view:addChild(self._table)
    self._table:setScale(view_size.height/750);

    local gap_x = 0;
    if view_size.width/view_size.height > 17/9 then
        gap_x = (500-240)*(view_size.height/750);
        self._table_old:setVisible(false);
    else 
        self._table:setVisible(false);
    end

    -- table part
    self._left_table_bg = control_tools.newImg({path=Layout3DTablePath.."table1.png"})
	self._left_table_bg:setScale(view_size.height/750);
    local l_left_table_width=self._left_table_bg:getSize().width*(view_size.height/750);
	
    local t_mid = control_tools.newImg({path=Layout3DTablePath.."table2.png"})
	
    self._right_table_bg = control_tools.newImg({path=Layout3DTablePath.."table3.png"})
	self._right_table_bg:setScale(view_size.height/750);
    local l_right_table_width=self._right_table_bg:getSize().width*(view_size.height/750);
	--local tt=self._left_table_bg:getSize().width
    local middle_width = view_size.width -l_left_table_width- l_right_table_width
    if middle_width < 640*(view_size.height/750) then middle_width = 640*(view_size.height/750) end

    self._left_table_bg:setPosition(ccp(-middle_width / 2 - l_left_table_width / 2 + gap_x/2, 0))
    self._right_table_bg:setPosition(ccp(middle_width / 2 + l_right_table_width / 2 - gap_x/2, 0))
    t_mid:setScaleX(middle_width / t_mid:getSize().width)
	t_mid:setScaleY(view_size.height/750)
    t_mid:setPosition(ccp(0,0))
    self._table_old:addChild(self._left_table_bg)
    self._table_old:addChild(t_mid)
    self._table_old:addChild(self._right_table_bg)

    self._left_table_mask = control_tools.newImg({path=Layout3DTablePath.."tablemask1.png",anchor=ccp(0,0.5)});
	self._left_table_mask:setScale(view_size.height/750);
	self._left_table_mask:setPosition(ccp(-850*view_size.height/750, 0))
	self._view:addChild(self._left_table_mask);
	self._right_table_mask = control_tools.newImg({path=Layout3DTablePath.."tablemask3.png",anchor=ccp(1,0.5)});	
	self._right_table_mask:setScale(view_size.height/750);
	self._right_table_mask:setPosition(ccp(850*view_size.height/750, 0))
    self._view:addChild(self._right_table_mask);
    -- direction indicator
    self._clock = control_tools.newImg({path="clock_bg.png", ctype=1})
    self._clock:setPosition(ccp(0,30))
    self._view:addChild(self._clock)
    self._clock.zi = {}
    self._clock.di = {}
    self._clock.light = {}
    self:setSelfDirection(TableFeng.none)

    -- time number
    self._clock.label=control_tools.newLabelBM({path=Layout3DFontPath.."num_jsq.fnt",str=""})
    self._clock.label:setPosition(ccp(0,0))
    self._clock.label:setVisible(true)
    self._clock:addChild(self._clock.label)

    self._clock:setVisible(false)
end

-- 显示或隐藏风位指示器
---@param bool_show bool
function table_3d_layout:showDirectionClock(bool_show)
    self._clock:setVisible(bool_show)
end

-- 设置自己的风位
---@param e_feng TableFeng
function table_3d_layout:setSelfDirection(e_feng)
    for i=1,4 do
        if self._clock.zi[i] then
            self._clock.zi[i]:removeFromParent()
        end
        if self._clock.di[i] then
            self._clock.di[i]:removeFromParent()
        end
        if self._clock.light[i] then
            self._clock.light[i]:removeFromParent()
        end
    end
    self._clock.zi[UserIndex.down],     self._clock.di[UserIndex.down], self._clock.light[UserIndex.down]   = self:loadDirection(e_feng, 11, 4)
    self._clock.zi[UserIndex.right],    self._clock.di[UserIndex.right],self._clock.light[UserIndex.right]  = self:loadDirection(e_feng, 21, 1)
    self._clock.zi[UserIndex.up],       self._clock.di[UserIndex.up],   self._clock.light[UserIndex.up]     = self:loadDirection(e_feng, 31, 3)
    self._clock.zi[UserIndex.left],     self._clock.di[UserIndex.left], self._clock.light[UserIndex.left]   = self:loadDirection(e_feng, 41, 2)
end

-- 点亮某个风位的指示灯
---@param e_index UserIndex
function table_3d_layout:lightDirection(e_index)
    if self._clock.di[e_index] then
        self._clock.di[e_index]:setVisible(true)
        local array=CCArray:create()
        array:addObject(CCFadeIn:create(1))
        array:addObject(CCFadeOut:create(1))
        local action=CCSequence:create(array)
        self._clock.di[e_index]:stopAllActions()
        self._clock.di[e_index]:runAction(CCRepeatForever:create(action))
    end
end

-- 熄灭所有指示灯
function table_3d_layout:turnOffAll()
    for k,v in pairs(self._clock.di) do
        v:stopAllActions()
        v:setOpacity(0)
    end
end

-- 点亮某个风位的灯光
---@param e_index UserIndex
function table_3d_layout:lightLightDirection(e_index)
    if self._clock.light[e_index] then
        self._clock.light[e_index]:setVisible(true)
        self._clock.light[e_index]:setOpacity(255)
        --local array=CCArray:create()
        --array:addObject(CCFadeIn:create(1))
        --array:addObject(CCFadeOut:create(1))
        --local action=CCSequence:create(array)
        --self._clock.light[e_index]:stopAllActions()
        --self._clock.light[e_index]:runAction(CCRepeatForever:create(action))
    end
end

-- 熄灭所有灯光
function table_3d_layout:turnOffAllLight()
    for k,v in pairs(self._clock.light) do
        v:setVisible(false)
        v:setOpacity(0)
    end
end

-- 显示或隐藏计时器
---@param bool_show bool
function table_3d_layout:showTime(bool_show)
    self._clock.label:setVisible(bool_show)
end

-- 设置计时器数字
---@param int_num int
---@param bool_auto_count_down bool @是否自动倒计时
function table_3d_layout:setTime(int_num, bool_auto_count_down)
    if int_num > 99 then int_num = 99 end
    if int_num < 0 then int_num = 0 end
    if int_num < 10 then
        self._clock.label:setText("0"..tostring(int_num))
    else
        self._clock.label:setText(tostring(int_num))
    end

    if bool_auto_count_down and bool_auto_count_down == true then
        self._time = int_num
        if self._time_schedule then
            CCDirector:sharedDirector():getScheduler():unscheduleScriptEntry(self._time_schedule)
            self._time_schedule=nil
        end
        self._time_schedule=CCDirector:sharedDirector():getScheduler():scheduleScriptFunc(function(dt) self:scheduleTimer(dt) end , 0.1,false)
    end
end

-- 设置计时器结束的callback
---@param cb_function function @function() 计时器结束以后会调用
function table_3d_layout:setTimeUpFunc(cb_function)
    if cb_function and type(cb_function) == "function" then
        self._timeup_function = cb_function
    end
end

-- 销毁
function table_3d_layout:destroy()
    if self._time_schedule then
        CCDirector:sharedDirector():getScheduler():unscheduleScriptEntry(self._time_schedule)
        self._time_schedule=nil
    end
    CCSpriteFrameCache:sharedSpriteFrameCache():removeSpriteFramesFromFile(Layout3DpListPath.."table.plist")
end

-- 私有方法

-- 自动计时器
---@param dt number @deltaT
function table_3d_layout:scheduleTimer(dt)
    self._time=self._time-dt
    if self._time<-0.1 then
        self:setTime(0)
        if self._time_schedule then
            CCDirector:sharedDirector():getScheduler():unscheduleScriptEntry(self._time_schedule)
            self._time_schedule=nil
            if self._timeup_function then
                self._timeup_function()
            end
        end
    elseif self._time<0 then
        self:setTime(0)
    else
        local num=math.modf(self._time)
        self:setTime(num+1)
    end
end

-- 设置单独风位
---@type private
---@param e_feng TableFeng
---@param num int
function table_3d_layout:loadDirection(e_feng, num, light_num)
    local zi,di,light
    local sub_path = ""

    if e_feng ~= TableFeng.none then
        zi = control_tools.newImg({path=sub_path.."clock_"..e_feng.."_"..num..".png", ctype=1})
        self._clock:addChild(zi)
    end
    di = control_tools.newImg({path=sub_path.."clock_"..e_feng.."_"..(num+1)..".png", ctype=1})
    di:setOpacity(0)
    self._clock:addChild(di)

    light = control_tools.newImg({path=sub_path.."clock_guang"..light_num..".png", ctype=1})
    light:setOpacity(0)
    light:setPosition(ccp(0, -9))
    self._clock:addChild(light)
    return zi,di,light
end


return table_3d_layout
